import string

# � character is used to represent unrecognizable characters in utf-8.
UNRECOGNIZABLE_CHARACTER = "�"
VALID_CHARACTERS = set(
    string.ascii_lowercase
    + string.ascii_uppercase
    + string.digits
    + UNRECOGNIZABLE_CHARACTER
    + string.punctuation
    + string.whitespace
    + "æøåÆØÅ"
)



from __future__ import annotations

import re
import warnings
from functools import lru_cache
from io import StringIO
from pathlib import Path

from yaml import safe_dump, safe_load

from cognite.powerops.utils.constants import UNRECOGNIZABLE_CHARACTER, VALID_CHARACTERS


def _validate(yaml_path: Path):
    if yaml_path.suffix not in {".yaml", ".yml"}:
        raise ValueError(f"File {yaml_path.name} not a valid yaml {yaml_path.suffix}")


@lru_cache
def load_yaml(yaml_path: Path, encoding="utf-8", clean_data: bool = False) -> dict:
    _validate(yaml_path)
    # The Windows Cpython implementation seems to guess if encoding is not explicitly set
    # This turns out to be a problem as it guesses wrong, which is not the case for Unix systems.
    data = Path(yaml_path).read_text(encoding=encoding)

    if clean_data and (invalid_characters := (set(data) - VALID_CHARACTERS)):
        data = re.sub(rf"[{'|'.join(invalid_characters)}]", UNRECOGNIZABLE_CHARACTER, data)
        warnings.warn(
            f"File {yaml_path.parent}/{yaml_path.name} contains invalid characters: {', '.join(invalid_characters)}"
        )
    return safe_load(StringIO(data))


# Having a single yaml dump function makes it easy to look up all places YaMLs are dumped.
# It also enables one place to set the default encoding.
def dump_yaml(data: dict, yaml_path: Path, encoding="utf-8"):
    _validate(yaml_path)
    with open(yaml_path, "w", encoding=encoding) as stream:
        safe_dump(data, stream)



import re
from uuid import uuid4


def replace_nordic_letters(input_string: str) -> str:
    return input_string
    # return (
    #     input_string.replace("æ", "a")
    #     .replace("Æ", "A")
    #     .replace("ø", "o")
    #     .replace("Ø", "O")
    #     .replace("å", "a")
    #     .replace("Å", "A")
    # )


def special_case_handle_gate_number(name: str) -> None:
    """Must handle any gate numbers above 1 if found in other sources than YAML"""
    # TODO: extend to handle special case if needed
    if re.search(pattern=r"L[2-9]", string=name):
        print_warning(f"Potential gate {name} not in YAML!")


def print_warning(s: str) -> None:
    """Adds some nice colors to the printed text :)"""
    print(f"\033[91m[WARNING] {s}\033[0m")


def unique_str() -> str:
    return str(uuid4())


def merge(*sources: dict) -> dict:
    """
    Merge multiple dicts:
     * Nested dicts are merged (hard type check).
     * When keys conflict, last source wins.
     * Order should be preserved. Hopefully.

    >>> merge({"a": 1, "c": 3}, {"b": 22, "c": 33})
    {'a': 1, 'c': 33, 'b': 22}

    >>> merge({"a": {"aa": 1}}, {"a": {"bb": 22}, "z": 33})
    {'a': {'aa': 1, 'bb': 22}, 'z': 33}
    """

    def _merge_two(val1, val2):
        if isinstance(val1, dict) and isinstance(val2, dict):
            return merge(val1, val2)
        else:
            return val2

    result = {}
    for source in sources:
        for key, val in source.items():
            if key in result:
                val = _merge_two(result[key], val)
            result[key] = val
    return result



from __future__ import annotations

import json
from typing import Any, ClassVar, Iterator, Literal, Optional

import pandas as pd
from cognite.client.data_classes import Sequence
from pydantic import BaseModel, validator

from cognite.powerops.data_classes.cdf_resource_collection import BootstrapResourceCollection, SequenceContent
from cognite.powerops.data_classes.common import AggregationMethod, RetrievalType
from cognite.powerops.data_classes.transformation import Transformation
from cognite.powerops.utils.common import print_warning

ATTRIBUTE_DEFAULT_AGGREGATION: dict[str, AggregationMethod | None] = {
    "gate.max_flow": AggregationMethod.min,
    "gate.min_flow": AggregationMethod.max,
    "gate.schedule_flag": AggregationMethod.first,
    "gate.schedule_m3s": AggregationMethod.first,
    "generator.committed_flag": AggregationMethod.first,
    "generator.committed_in": None,
    "generator.initial_state": None,
    "generator.maintenance_flag": AggregationMethod.max,
    "generator.max_p_constr": AggregationMethod.min,
    "generator.min_p_constr": AggregationMethod.max,
    "generator.production_schedule_flag": AggregationMethod.first,
    "market.buy_price": AggregationMethod.mean,
    "market.sale_price": AggregationMethod.mean,
    "plant.feeding_fee": AggregationMethod.mean,
    "plant.maintenance_flag": AggregationMethod.max,
    "plant.max_p_constr": AggregationMethod.min,
    "plant.max_q_constr": AggregationMethod.min,
    "plant.min_p_constr": AggregationMethod.max,
    "plant.mip_flag": AggregationMethod.first,
    "plant.production_schedule": AggregationMethod.first,
    "plant.production_schedule_flag": AggregationMethod.first,
    "reserve_group.rr_up_obligation": AggregationMethod.max,
    "reserve_group.rr_down_obligation": AggregationMethod.max,
    "reservoir.inflow": AggregationMethod.mean,
    "reservoir.max_vol_constr": AggregationMethod.min,
    "reservoir.min_vol_constr": AggregationMethod.max,
    "reservoir.start_head": None,
    "reservoir.tactical_cost_max": AggregationMethod.mean,
    "reservoir.tactical_cost_min": AggregationMethod.mean,
    "reservoir.tactical_limit_max": AggregationMethod.mean,
    "reservoir.tactical_limit_min": AggregationMethod.mean,
}


class TimeSeriesMappingEntry(BaseModel):
    object_type: str
    object_name: str
    attribute_name: str
    time_series_external_id: Optional[str] = None
    transformations: Optional[list[Transformation]] = None
    retrieve: Optional[RetrievalType] = None
    aggregation: Optional[AggregationMethod] = None

    @validator("aggregation", pre=True)
    def to_enum(cls, value):
        return AggregationMethod[value] if isinstance(value, str) else value

    @validator("aggregation", always=True)
    def set_default(cls, value, values):
        if value is not None:
            return value
        # TODO: do we want default `None` here or raise error?
        return ATTRIBUTE_DEFAULT_AGGREGATION.get(f"{values.get('object_type')}.{values.get('attribute_name')}")

    @validator("retrieve", pre=True)
    def to_retrival_enum(cls, value):
        return RetrievalType[value] if isinstance(value, str) else value

    @property
    def shop_model_path(self) -> str:
        return f"{self.object_type}.{self.object_name}.{self.attribute_name}"

    def _transformations_to_strings(self, max_cols: int) -> list[str]:
        if self.transformations:
            # ensure_ascii=False to allow æøåÆØÅ
            transformation_string = json.dumps([t.to_dict() for t in self.transformations], ensure_ascii=False)
        else:
            transformation_string = json.dumps([])

        max_chars = 255
        return [transformation_string[i * max_chars : (i + 1) * max_chars] for i in range(max_cols)]

    def to_sequence_row(self, max_transformation_cols: int) -> list[str | float]:
        return [
            self.shop_model_path,
            self.time_series_external_id or float("nan"),
            *self._transformations_to_strings(max_cols=max_transformation_cols),
            self.retrieve.name if self.retrieve else float("nan"),
            self.aggregation.name if self.aggregation else float("nan"),
        ]


class TimeSeriesMapping(BaseModel):
    rows: list[TimeSeriesMappingEntry] = []
    columns: ClassVar[list[str]] = [
        "shop_model_path",
        "time_series_external_id",
        "transformations",
        "transformations1",
        "transformations2",
        "transformations3",
        "retrieve",
        "aggregation",
    ]

    @property
    def transformations_cols(self) -> list[str]:
        return [col for col in self.columns if col.startswith("transformations")]

    def __iter__(self) -> Iterator[TimeSeriesMappingEntry]:  # type: ignore [override]
        yield from self.rows

    def __len__(self) -> int:
        return len(self.rows)

    def __add__(self, other: "TimeSeriesMapping") -> "TimeSeriesMapping":
        return TimeSeriesMapping(rows=self.rows + other.rows)

    def append(self, element: TimeSeriesMappingEntry) -> None:
        self.rows.append(element)

    def extend(self, other: "TimeSeriesMapping") -> None:
        self.rows.extend(other.rows)

    @property
    def column_definitions(self) -> list[dict]:
        return [{"valueType": "STRING", "externalId": col} for col in self.columns]

    def to_sequence_rows(self) -> dict[int, list[str | float]]:
        return {
            i: row.to_sequence_row(max_transformation_cols=len(self.transformations_cols))
            for i, row in enumerate(self.rows)
        }

    def to_dataframe(self) -> pd.DataFrame:
        rows = [row.to_sequence_row(max_transformation_cols=len(self.transformations_cols)) for row in self.rows]
        return pd.DataFrame(data=rows, columns=self.columns)

    def dumps(self) -> dict[str, Any]:
        rows = []
        for row in self.rows:
            row_raw = row.dict(exclude={"aggregation", "retrieve", "transformations"})
            if row.aggregation:
                row_raw["aggregation"] = row.aggregation.name
            if row.retrieve:
                row_raw["retrieve"] = row.retrieve.name
            if row.transformations:
                row_raw["transformations"] = [
                    {
                        **transformation.dict(exclude={"transformation"}),
                        "transformation": transformation.transformation.name,
                    }
                    for transformation in row.transformations
                ]
            rows.append(row_raw)
        return {"rows": rows}


def write_mapping_to_sequence(
    mapping: TimeSeriesMapping,
    watercourse: str,
    mapping_type: Literal["base_mapping", "incremental_mapping", "rkom_incremental_mapping"],
    price_scenario_name: str = "",  # Required for incremental mapping
    config_name: str = "",  # Required for incremental mapping
    reserve_volume: int = -1,  # Required for rkom
) -> BootstrapResourceCollection:
    if mapping_type not in ["base_mapping", "incremental_mapping", "rkom_incremental_mapping"]:
        raise ValueError(f"Unrecognized mapping type: {mapping_type}")

    if mapping_type == "incremental_mapping" and (not price_scenario_name or not config_name):
        raise ValueError(
            "Both scenario_name and config_name must be specified when mapping_type='incremental_mapping'!"
        )

    if mapping_type == "rkom_incremental_mapping" and reserve_volume < 0:
        raise ValueError("'reserve_volume' must be specified when mapping_type='rkom_incremental_mapping'")

    metadata = {
        "shop:watercourse": watercourse,
        "shop:type": mapping_type,
    }

    if mapping_type == "base_mapping":
        external_id = f"SHOP_{watercourse}_base_mapping"
        name = external_id.replace("_", " ")

    elif mapping_type == "incremental_mapping":
        external_id = f"SHOP_{watercourse}_incremental_mapping_{config_name}_{price_scenario_name}"
        name = external_id.replace("_", " ")
        metadata["bid:scenario_name"] = price_scenario_name

    elif mapping_type == "rkom_incremental_mapping":
        external_id = f"SHOP_{watercourse}_incremental_mapping_{config_name}_{price_scenario_name}_{reserve_volume}MW"
        name = f"{watercourse} {price_scenario_name} {reserve_volume} MW"
        metadata["bid:scenario_name"] = price_scenario_name
        metadata["bid:reserve_volume"] = str(reserve_volume)

    else:
        raise ValueError(f"Unrecognized mapping type: {mapping_type}")

    sequence = Sequence(
        name=name,
        external_id=external_id,
        description="Mapping between SHOP paths and CDF TimeSeries",
        columns=mapping.column_definitions,
        metadata=metadata,
    )
    bootstrap_resource_collection = BootstrapResourceCollection()
    bootstrap_resource_collection.add(sequence)
    sequence_dataframe = mapping.to_dataframe()
    if not sequence_dataframe.empty:
        bootstrap_resource_collection.add(
            SequenceContent(sequence_external_id=sequence.external_id, data=sequence_dataframe)
        )
    else:
        print_warning("Time series mapping is empty! No sequence rows to write!")
    return bootstrap_resource_collection



from abc import ABC, abstractmethod
from datetime import datetime, timedelta
from typing import Any, Dict, Optional, Union

import numpy as np
import pandas as pd
from cognite.client import CogniteClient

from cognite.powerops.preprocessor import knockoff_logging as logging
from cognite.powerops.preprocessor.exceptions import CogShopError
from cognite.powerops.preprocessor.utils import ms_to_datetime, retrieve_range

logger = logging.getLogger(__name__)

RelativeDatapoints = Dict[str, float]


def _relative_datapoints_to_series(
    relative_datapoints: RelativeDatapoints, start_time: datetime, shift_minutes: int = 0
) -> pd.Series:
    """Converts 'relative datapoints' to a pandas Series

    Example:
        >>> start_time = datetime(2000, 1, 1, 12)
        >>> relative_datapoints = {
        ...     0: 42,
        ...     60: 420,
        ...     1440: 4200,
        ... }
        >>> _relative_datapoints_to_series(relative_datapoints, start_time)
        2000-01-01 12:00:00      42
        2000-01-01 13:00:00     420
        2000-01-02 12:00:00    4200
        dtype: int64
    """
    return pd.Series(
        relative_datapoints.values(),
        index=[
            start_time + timedelta(minutes=int(minute_offset) + shift_minutes) for minute_offset in relative_datapoints
        ],
    )


# TODO: can we reduce the number of parameters in `apply`?
# ? SimpleTransformation.apply(datatapoints) and ComplexTransformation.apply(<..all params..>)
class Transformation(ABC):
    def __init__(self, kwargs: Optional[dict] = None) -> None:
        pass

    @abstractmethod
    def apply(
        self,
        *_: Any,
        datapoints: pd.Series,
        model: dict,
        object_type: str,
        instance: Union[str, int],
        start: int,
        end: int,
        client: CogniteClient,
        shift_minutes: int = 0,
    ) -> pd.Series:
        ...


class Add(Transformation):
    def __init__(self, kwargs: dict) -> None:
        self.value = kwargs["value"]

    def apply(self, *_: Any, datapoints: pd.Series, **__: Any) -> pd.Series:
        return datapoints + self.value


class Multiply(Transformation):
    def __init__(self, kwargs: dict) -> None:
        self.value = kwargs["value"]

    def apply(self, *_: Any, datapoints: pd.Series, **__: Any) -> pd.Series:
        return datapoints * self.value


class StaticValues(Transformation):
    def __init__(self, kwargs: dict) -> None:
        self.relative_datapoints = kwargs

    def apply(self, *_: Any, start: int, shift_minutes: int = 0, **__: Any) -> pd.Series:
        return _relative_datapoints_to_series(self.relative_datapoints, ms_to_datetime(start), shift_minutes)


class ToBool(Transformation):
    @staticmethod
    def apply(*_: Any, datapoints: pd.Series, **__: Any) -> pd.Series:
        return (datapoints > 0).astype(int)


class ZeroIfNotOne(Transformation):
    @staticmethod
    def apply(*_: Any, datapoints: pd.Series, **__: Any) -> pd.Series:
        return (datapoints == 1).astype(int)


class OneIfTwo(Transformation):
    @staticmethod
    def apply(*_: Any, datapoints: pd.Series, **__: Any) -> pd.Series:
        return (datapoints == 2).astype(int)


class HeightToVolume(Transformation):
    @staticmethod
    def height_to_volume(datapoints, heights, volumes) -> pd.Series:
        def interpolate(height: float) -> float:
            """Height to volume"""
            if height < min(heights) or max(heights) < height:
                logger.warning(f"Outside interpoaltion bounds [{min(heights)}, {max(heights)}]. Got {height}.")
            return float(np.interp(height, heights, volumes))

        return datapoints.map(interpolate)

    def apply(
        self,
        *_: Any,
        datapoints: pd.Series,
        model: dict,
        object_type: str,
        instance: Union[str, int],
        **__: Any,
    ) -> pd.Series:
        volumes = model[object_type][instance]["vol_head"]["x"]
        heights = model[object_type][instance]["vol_head"]["y"]
        return self.height_to_volume(datapoints, heights, volumes)


class DoNothing(Transformation):
    @staticmethod
    def apply(*_: Any, datapoints: pd.Series, **__: Any) -> pd.Series:
        return datapoints


class AddFromOffset(Transformation):
    def __init__(self, kwargs) -> None:
        self.relative_datapoints = kwargs

    def apply(
        self,
        *_: Any,
        datapoints: pd.Series,
        shift_minutes: int = 0,
        **__: Any,
    ) -> pd.Series:
        first_timestamp = min(datapoints.index)
        non_relative_datapoints = _relative_datapoints_to_series(
            self.relative_datapoints, first_timestamp, shift_minutes
        )
        union_index = datapoints.index.union(non_relative_datapoints.index)
        # fillna(0) since we are adding
        non_relative_datapoints = non_relative_datapoints.reindex(union_index).ffill().fillna(0)
        # If a timestamp does not exist in the original datapoints we add that timestamp with a forwardfilled value
        return datapoints.reindex(union_index).ffill() + non_relative_datapoints


class MultiplyFromOffset(Transformation):
    def __init__(self, kwargs) -> None:
        self.relative_datapoints = kwargs

    def apply(
        self,
        *_: Any,
        datapoints: pd.Series,
        shift_minutes: int = 0,
        **__: Any,
    ) -> pd.Series:
        first_timestamp = min(datapoints.index)
        non_relative_datapoints = _relative_datapoints_to_series(
            self.relative_datapoints, first_timestamp, shift_minutes
        )
        union_index = datapoints.index.union(non_relative_datapoints.index)
        # fillna(1) since we are multiplying
        non_relative_datapoints = non_relative_datapoints.reindex(union_index).ffill().fillna(1)
        # If a timestamp does not exist in the original datapoints we add that timestamp with a forwardfilled value
        return datapoints.reindex(union_index).ffill() * non_relative_datapoints


class AddWaterInTransit(Transformation):
    def __init__(self, kwargs) -> None:
        self.kwargs = kwargs

    @staticmethod
    def add_water_in_transit(
        inflow: pd.Series,
        discharge: pd.Series,
        shape: Dict[int, float],
        start: datetime,
        end: datetime,
    ) -> pd.Series:
        """Adds water in transit (previously discharged water) to the inflow time series.

        Args:
            shape (Dict[int, float]): Description of "how much water is delayed by what amount of time".
                e.g.
                shape = {
                    0: 0,       # 0% instantly
                    120: 0.5,   # 50% delayed for 2h
                    1320: 0.5,  # 50% delayed for 22h
                }

        """
        # Forward fill discharge for all (hour) timestamps until start
        one_hour = pd.Timedelta("1h")
        if start - one_hour not in discharge.index:
            discharge[start - one_hour] = np.NaN
        discharge = discharge.resample(one_hour).ffill().ffill()

        # Make sure inflow has values for all (hour) timestamps up until end (exclusive)
        if end not in inflow.index:
            inflow[end] = np.NaN
        inflow = inflow.resample(one_hour).ffill().ffill()

        for delay, water_percentage in shape.items():
            delayed_discharge = discharge.shift(delay, freq="min") * water_percentage
            inflow = inflow.add(delayed_discharge, fill_value=0)

        # Only include delayed discharge that will affect the selected time range
        between_start_and_end = (start <= inflow.index) & (inflow.index < end)
        return inflow.loc[between_start_and_end]

    @staticmethod
    def get_shape(model: Dict, object_type: str, object_name: str) -> Dict[int, float]:
        """object_type must be plant or gate"""
        gate_or_plant = model[object_type][object_name]  # Get description of gate/plant

        # TODO: HANDLE REF VALUES

        # Get shape and time_delay values as Dict[delay,water_percentage]
        if "shape_discharge" in gate_or_plant:
            shape = dict(zip(gate_or_plant["shape_discharge"]["x"], gate_or_plant["shape_discharge"]["y"]))
        elif "time_delay" in gate_or_plant:
            shape = {gate_or_plant["time_delay"]: 1}
        else:
            raise ValueError(f"{object_type}.{object_name} does not have a shape_discharge or time_delay attribute")

        return shape

    def apply(
        self,
        *_: Any,
        datapoints: pd.Series,
        model: dict,
        start: int,
        end: int,
        client: CogniteClient,
        **__: Any,
    ) -> pd.Series:
        if datapoints.empty:
            return datapoints

        if "gate_name" in self.kwargs:
            shape = self.get_shape(model=model, object_type="gate", object_name=self.kwargs["gate_name"])
        else:  # ! Assumes that this is the only other option
            shape = self.get_shape(model=model, object_type="plant", object_name=self.kwargs["plant_name"])

        longest_delay = max(shape)  # longest delay in minutes
        longest_delay_ms = 60 * 1000 * longest_delay  # longest delay in milliseconds

        external_id = self.kwargs["external_id"]
        discharge = retrieve_range(  # Get discharge datapoints from time-series
            client=client,
            external_ids=[external_id],
            start=start - longest_delay_ms,  # Shift start time based on longest delay
            end=start,
        )[external_id]

        if discharge.empty:
            logger.warning("Cannot add 'water in transit' - did not get any 'discharge' datapoints!")
            return datapoints

        return self.add_water_in_transit(
            inflow=datapoints,
            discharge=discharge,
            shape=shape,
            start=ms_to_datetime(start),
            end=ms_to_datetime(end),
        )


_TRANSFORMATIONS = {
    "ADD": Add,
    "ADD_FROM_OFFSET": AddFromOffset,
    "DYNAMIC_ADD_FROM_OFFSET": AddFromOffset,
    "ADD_WATER_IN_TRANSIT": AddWaterInTransit,
    "MULTIPLY": Multiply,
    "MULTIPLY_FROM_OFFSET": MultiplyFromOffset,
    "TO_BOOL": ToBool,
    "ZERO_IF_NOT_ONE": ZeroIfNotOne,
    "GATE_SCHEDULE_FLAG_VALUE_MAPPING": ZeroIfNotOne,
    "GENERATOR_PRODUCTION_SCHEDULE_FLAG_VALUE_MAPPING": ZeroIfNotOne,
    "PLANT_PRODUCTION_SCHEDULE_FLAG_VALUE_MAPPING": ZeroIfNotOne,
    "ONE_IF_TWO": OneIfTwo,
    "STATIC": StaticValues,
    "DYNAMIC_STATIC": StaticValues,
    "RESERVOIR_LEVEL_TO_VOLUME": HeightToVolume,
    "DO_NOTHING": DoNothing,
    "GATE_OPENING_METER_TO_PERCENT": DoNothing,
}


def transformation_factory(transformation_type: str, kwargs: dict) -> Transformation:
    try:
        return _TRANSFORMATIONS[transformation_type](kwargs)
    except KeyError as e:
        raise CogShopError(f"{transformation_type} not recognized as a 'Transformation'!") from e



import logging

import tomli_w

from cognite.powerops.settings.data_classes import Settings

__all__ = ["settings"]

settings = Settings()

if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    # This is just a convenience for development, prints out all settings as TOML.
    # TOML has no support for `None`, filter it out:
    _dump = tomli_w.dumps(
        {
            group_key: {key: val for key, val in group_val.items() if val is not None}
            for group_key, group_val in settings.dict().items()
        }
    )
    print(_dump)



from __future__ import annotations

from typing import Sequence, Type

from cognite.client import CogniteClient
from cognite.client._constants import LIST_LIMIT_DEFAULT
from cognite.client.data_classes import Asset
from pydantic import BaseModel

from cognite.powerops.config import (
    BenchmarkingConfig,
    BidProcessConfig,
    MarketConfig,
    RKOMBidCombinationConfig,
    RKOMBidProcessConfig,
    Watercourse,
)
from cognite.powerops.data_classes.asset_lists import (
    BenchmarkingList,
    BidProcessList,
    GeneratorList,
    MarketConfigsList,
    PlantList,
    PriceAreaList,
    ReservoirList,
    RKOMBidCombinationList,
    RKOMBidList,
    WatercourseList,
)
from cognite.powerops.data_classes.core import T_AssetResourceList
from cognite.powerops.data_classes.generator import Generator
from cognite.powerops.data_classes.plant import Plant
from cognite.powerops.data_classes.price_area import PriceArea
from cognite.powerops.data_classes.reservoirs import Reservoir


def unpack_asset(asset: Asset) -> dict:
    unpacked = asset.dump(camel_case=False)
    if "metadata" in unpacked:
        metadata = unpacked.pop("metadata")
        for key, value in metadata.items():
            unpacked[key.replace(":", "_")] = value

    return unpacked


class AssetAPI:
    def __init__(
        self,
        client: CogniteClient,
        read_dataset: str,
        write_dataset: str,
        parent_external_id: str,
        external_id_prefix: str,
        class_type: Type[BaseModel],
        class_list: Type[T_AssetResourceList],
    ):
        self._client = client
        self._external_id_prefix = external_id_prefix
        self._parent_external_id = parent_external_id
        self._read_dataset = read_dataset
        self._write_dataset = write_dataset
        self._class_type = class_type
        self._class_list = class_list

    def _to_list_type(self, assets: Sequence[Asset]) -> T_AssetResourceList:
        return self._class_list([self._class_type(**unpack_asset(asset)) for asset in assets])

    def list(self, limit: int = LIST_LIMIT_DEFAULT) -> T_AssetResourceList:
        assets = self._client.assets.list(
            parent_external_ids=[self._parent_external_id],
            data_set_external_ids=[self._read_dataset, self._write_dataset],
            limit=limit,
        )
        return self._to_list_type(assets)

    def retrieve(self, name: str | list[str], ignore_unknown_ids: bool = False) -> BaseModel | T_AssetResourceList:
        names = name if isinstance(name, list) else [name]
        assets = self._client.assets.retrieve_multiple(
            external_ids=[f"{self._external_id_prefix}{name}" for name in names], ignore_unknown_ids=ignore_unknown_ids
        )

        if len(assets) == 1:
            return self._class_type(**unpack_asset(assets[0]))
        return self._to_list_type(assets)


class WatercourseAPI(AssetAPI):
    """Manage watercourses. Changes are directly applied to CDF."""

    def __init__(self, client: CogniteClient, read_dataset: str, write_dataset: str):
        super().__init__(
            client, read_dataset, write_dataset, "watercourses", "watercourse_", Watercourse, WatercourseList
        )

    #
    # def copy(self, watercourse: Watercourse, name: str) -> Watercourse:
    #     """Create a copy of an existing watercourse, with a new name."""


class PlantAPI(AssetAPI):
    def __init__(self, client: CogniteClient, read_dataset: str, write_dataset: str):
        super().__init__(client, read_dataset, write_dataset, "plants", "plant_", Plant, PlantList)


class ReservoirsAPI(AssetAPI):
    def __init__(self, client: CogniteClient, read_dataset: str, write_dataset: str):
        super().__init__(client, read_dataset, write_dataset, "reservoirs", "reservoir_", Reservoir, ReservoirList)


class GeneratorsAPI(AssetAPI):
    def __init__(self, client: CogniteClient, read_dataset: str, write_dataset: str):
        super().__init__(client, read_dataset, write_dataset, "generators", "generator_", Generator, GeneratorList)


class PriceAreasAPI(AssetAPI):
    def __init__(self, client: CogniteClient, read_dataset: str, write_dataset: str):
        super().__init__(client, read_dataset, write_dataset, "price_areas", "price_area_", PriceArea, PriceAreaList)


class BidConfigurationsAPI(AssetAPI):
    def __init__(self, client: CogniteClient, read_dataset: str, write_dataset: str):
        super().__init__(
            client,
            read_dataset,
            write_dataset,
            "bid_process_configurations",
            "POWEROPS_bid_process_configuration_",
            BidProcessConfig,
            BidProcessList,
        )


class RKOMBidConfigurationsAPI(AssetAPI):
    def __init__(self, client: CogniteClient, read_dataset: str, write_dataset: str):
        super().__init__(
            client,
            read_dataset,
            write_dataset,
            "rkom_bid_process_configurations",
            "POWEROPS_",
            RKOMBidProcessConfig,
            RKOMBidList,
        )


class BenchmarkingConfigurationsAPI(AssetAPI):
    def __init__(self, client: CogniteClient, read_dataset: str, write_dataset: str):
        super().__init__(
            client,
            read_dataset,
            write_dataset,
            "benchmarking_configurations",
            "POWEROPS_",
            BenchmarkingConfig,
            BenchmarkingList,
        )


class MarketConfigurationsAPI(AssetAPI):
    def __init__(self, client: CogniteClient, read_dataset: str, write_dataset: str):
        super().__init__(
            client,
            read_dataset,
            write_dataset,
            "market_configurations",
            "market_configuration_",
            MarketConfig,
            MarketConfigsList,
        )


class RKOMBidCombinationConfiguration(AssetAPI):
    def __init__(self, client: CogniteClient, read_dataset: str, write_dataset: str):
        super().__init__(
            client,
            read_dataset,
            write_dataset,
            "rkom_bid_combination_configurations",
            "RKOM_bid_combination_configuration_",
            RKOMBidCombinationConfig,
            RKOMBidCombinationList,
        )



from __future__ import annotations

import tempfile
from functools import cached_property
from typing import List, Literal, Optional, TypedDict, Union

import arrow
from cognite.client import CogniteClient
from cognite.client.data_classes import FileMetadata
from pydantic import BaseModel, root_validator

from cognite.powerops.preprocessor import knockoff_logging as logging
from cognite.powerops.preprocessor.data_classes import CogShopCase, CogShopConfig
from cognite.powerops.preprocessor.data_classes.time_series_mapping import TimeSeriesMapping
from cognite.powerops.preprocessor.exceptions import CogReaderError
from cognite.powerops.preprocessor.utils import (
    ShopMetadata,
    arrow_to_ms,
    download_file,
    log_and_reraise,
    now,
    retrieve_yaml_file,
)

from .get_fdm_data import Case, get_case

logger = logging.getLogger(__name__)


class CogShopFileDict(TypedDict):
    external_id: str
    file_type: str


class SortBy(BaseModel):
    metadata_key: Optional[str]
    file_attribute: Optional[Literal["last_updated_time", "created_time", "uploaded_time"]]

    @root_validator
    def only_one_sorting_method(cls, values):
        if values.get("metadata_key") and values.get("file_attribute"):
            raise ValueError("Both metadata_key and file_attribute to sort by cannot be set for file")
        return values


class CogShopFile(BaseModel):
    label: Optional[str]
    pick: Optional[Literal["latest_before", "exact", "latest"]]
    sort_by: Optional[SortBy]
    external_id_prefix: Optional[str]
    external_id: Optional[str]
    file_type: Literal["ascii", "yaml"]

    class Config:
        validate_all = True

    @root_validator
    def external_id_or_prefix_set(cls, values):
        if not values.get("external_id") and not values.get("external_id_prefix"):
            raise ValueError("Either external_id or external_id_prefix must be set")
        return values

    def _find_file_latest_before(self, files: list[FileMetadata], starttime_ms: float) -> Optional[FileMetadata]:
        # Select file that is closest in time before starttime
        valid_files = []

        best_diff = float("inf")
        closest_file = None

        for this_file in files:
            try:
                if self.sort_by:
                    if key := self.sort_by.metadata_key:
                        if updated_at := this_file.metadata.get(key):
                            updated_at_ms = arrow_to_ms(arrow.get(updated_at))
                        else:
                            logger.warning(
                                f"Provided metadata field: {key} not in file metadata for {this_file.external_id} - "
                                f"skipping file"
                            )
                            continue
                    elif attr := self.sort_by.file_attribute:
                        updated_at_ms = getattr(this_file, attr)
                else:
                    logger.debug("No file selection method chosen. Will use last_updated_time attribute for file.")
                    updated_at_ms = getattr(this_file, "last_updated_time")

                # Assume datetime string after last "_"
                this_diff = starttime_ms - updated_at_ms
                if this_diff >= 0:
                    valid_files.append(this_file)
                    if this_diff < best_diff:
                        logger.debug(f"Cutfile {this_file.external_id} is closer to starttime.")
                        best_diff = this_diff
                        closest_file = this_file

            except (arrow.parser.ParserError, TypeError) as e:
                logger.warning(f"Failed to parse date for '{this_file.external_id}': {e}")

        if not closest_file and valid_files:
            logger.warning(f"Could not find a cut file with a valid datetime - using {valid_files[0].external_id}")
            return valid_files[0]
        elif not valid_files:
            return None

        return closest_file

    def get_file_dict(self, client: CogniteClient, starttime_ms: float) -> CogShopFileDict:
        if self.external_id:
            return {"external_id": self.external_id, "file_type": self.file_type}
        elif self.external_id_prefix and self.pick == "latest_before":
            files = client.files.list(external_id_prefix=self.external_id_prefix, limit=None)
            if closest_file := self._find_file_latest_before(files, starttime_ms):
                return {"external_id": closest_file.external_id, "file_type": self.file_type}
        elif self.external_id_prefix and self.pick == "latest":
            files = client.files.list(external_id_prefix=self.external_id_prefix, limit=None)
            if closest_file := self._find_file_latest_before(files, now()):
                return {"external_id": closest_file.external_id, "file_type": self.file_type}
        logger.warning("File is not accompanied by selection method. Returning empty file dictionary")
        return CogShopFileDict(external_id="", file_type="")


class CogShopFilesConfig(BaseModel):
    file_load_sequence: list[CogShopFile]

    @classmethod
    def from_cdf(cls, client: CogniteClient, file_ex_id: str) -> "CogShopFilesConfig":
        config = retrieve_yaml_file(client, file_ex_id)
        return cls(**config)

    def to_cog_shop_file_list(self, client: CogniteClient, starttime_ms: float) -> list[CogShopFileDict]:
        return [file.get_file_dict(client, starttime_ms) for file in self.file_load_sequence]


class CogReader:
    """Read data from CDF and preprocess for SHOP"""

    def __init__(
        self,
        client: CogniteClient,
        fdm_space_external_id: str,
        fdm_case_external_id: str,
        fdm_model_external_id: Optional[str] = None,
        fdm_model_version: Optional[int] = None,
    ) -> None:
        self.client: CogniteClient = client

        self.fdm_space_external_id = fdm_space_external_id
        self.fdm_case_external_id = fdm_case_external_id
        self.fdm_model_external_id = fdm_model_external_id
        self.fdm_model_version = fdm_model_version

        self._tmp_dir = tempfile.TemporaryDirectory()
        self.cog_shop_files_config: CogShopFilesConfig
        logger.info(f"{self.__class__.__name__} initialized.")

    @property
    def file_external_id_prefix(self) -> str:
        return f"SHOP_{self.cog_shop_config.watercourse}_"

    def _set_cog_shop_files_config(self) -> None:
        if not (
            file := self.client.files.retrieve(
                external_id=f"SHOP_{self.cog_shop_config.watercourse}_cog_shop_files_config"
            )
        ):
            raise CogReaderError(
                f"No cog_shop_files_config for watercourse {self.cog_shop_config.watercourse} in CDF. "
                f"Ensure that file exists"
            )
        logger.debug("Attempting to download cog shop files config from CDF.")
        self.cog_shop_files_config = CogShopFilesConfig.from_cdf(self.client, file.external_id)
        logger.info("Cog Shop file list config successfully initialised")

    @property
    def fdm_case(self) -> Case:
        try:
            return self._fdm_case
        except AttributeError:
            raise RuntimeError("Expected to run `retrieve_case()` method first.") from None

    def retrieve_case(self) -> None:
        self._fdm_case = get_case(
            self.client,
            self.fdm_space_external_id,
            self.fdm_case_external_id,
            model_extenal_id=self.fdm_model_external_id,
            model_version=self.fdm_model_version,
        )

    @cached_property
    def cog_shop_config(self) -> CogShopConfig:
        config = CogShopConfig(
            watercourse=self.fdm_case.scenario.model_template.watercourse,
            starttime=self.fdm_case.start_time,
            endtime=self.fdm_case.end_time,
        )
        logger.debug(f"{config} loaded from event metadata.")
        return config

    @cached_property
    def cog_shop_case(self) -> CogShopCase:
        """
        Download and parse model yaml file from cdf and merge with commands.
        Instantiate a CogShopCase from the resulting dict.
        """
        case_dict = retrieve_yaml_file(
            self.client,
            file_external_id=self.fdm_case.scenario.model_template.model.file_external_id,
        )
        case_dict["commands"] = self.fdm_case.scenario.commands.commands

        cog_shop_case = CogShopCase.from_dict(case_dict)
        logger.debug("SHOP case loaded.")
        return cog_shop_case

    @cached_property
    def base_mapping(self) -> List[TimeSeriesMapping]:
        mappings = self.fdm_case.scenario.model_template.base_mappings.items
        return [TimeSeriesMapping.from_mapping_model(m) for m in mappings]

    @cached_property
    def incremental_mapping(self) -> List[TimeSeriesMapping]:
        if (mo := self.fdm_case.scenario.mappings_override) is not None:
            return [TimeSeriesMapping.from_mapping_model(m) for m in mo.items]
        return []

    def get_cog_shop_file_list(self) -> list[CogShopFileDict]:
        return self.cog_shop_files_config.to_cog_shop_file_list(self.client, self.cog_shop_config.starttime_ms)

    @staticmethod
    def file_metadata_to_dict(file_metadata) -> dict[str, Union[str, int]]:
        md = {"id": file_metadata.id}
        if (xid := file_metadata.external_id) is not None:
            md["external_id"] = xid
        if (n := file_metadata.name) is not None:
            md["name"] = n
        if (dsid := file_metadata.data_set_id) is not None:
            md["data_set_id"] = dsid
        return md

    @cached_property
    def license_file_path(self) -> Optional[str]:
        """Read SHOP license file from CDF, store locally"""
        logger.debug("Looking for license file.")
        files = self.client.files.list(metadata=ShopMetadata(type="license"))
        if not files:
            logger.warning("Did not find any license file in CDF.")
            return None
        if len(files) > 1:
            logger.warning(f"Found more than one license file: {files}. Will use {files[0]}.")
        file_name = files[0].name
        if file_name != "SHOP_license.dat":
            raise CogReaderError(f"License file has to be named 'SHOP_license.dat'! Given file is named '{file_name}'.")
        download_file(client=self.client, file=files[0], download_directory=self._tmp_dir.name)
        return self._tmp_dir.name  # path to folder with license file instead of path to file

    @log_and_reraise(CogReaderError)
    def _replace_model_time_series(self) -> None:
        start = self.cog_shop_config.starttime_ms
        end = self.cog_shop_config.endtime_ms
        logger.debug("Applying 'base mapping'")
        self.cog_shop_case.replace_time_series(client=self.client, start=start, end=end, mappings=self.base_mapping)
        if self.incremental_mapping:
            logger.debug("Applying 'incremental mapping'")
            self.cog_shop_case.replace_time_series(
                client=self.client,
                start=start,
                end=end,
                mappings=self.incremental_mapping,
                dynamic_minute_offset=self.cog_shop_config.dynamic_minute_offset,
            )

    def run(self) -> CogReader:
        self.retrieve_case()

        # If specified in CogShopConfig (overrides previously set time specification)
        if self.cog_shop_config.timeresolution:
            self.cog_shop_case.set_time(
                starttime=self.cog_shop_config.starttime,
                endtime=self.cog_shop_config.endtime,
                timeresolution=self.cog_shop_config.timeresolution,
            )

        # Use defaults if time specification not already set
        if self.cog_shop_case.time is None:
            self.cog_shop_case.set_default_time(
                starttime=self.cog_shop_config.starttime, endtime=self.cog_shop_config.endtime
            )

        # TODO: consider if needed or commands should be required
        if self.cog_shop_case.commands is None:
            self.cog_shop_case._set_some_commands()

        self._replace_model_time_series()

        self._set_cog_shop_files_config()

        return self



from functools import cached_property

from cognite.client import ClientConfig, CogniteClient

from cognite.powerops.client.api.asset_apis import (
    BenchmarkingConfigurationsAPI,
    BidConfigurationsAPI,
    GeneratorsAPI,
    MarketConfigurationsAPI,
    PlantAPI,
    PriceAreasAPI,
    ReservoirsAPI,
    RKOMBidCombinationConfiguration,
    RKOMBidConfigurationsAPI,
    WatercourseAPI,
)
from cognite.powerops.client.api.shop_api import ShopAPI
from cognite.powerops.client.config_client import ConfigurationClient
from cognite.powerops.settings import settings
from cognite.powerops.utils.cdf_auth import get_client_config, get_cognite_client
from cognite.powerops.utils.cdf_utils import retrieve_dataset


class ConfigurationsClient:
    def __init__(self, read_dataset: str, write_dataset: str, client: CogniteClient):
        self.bids = BidConfigurationsAPI(client, read_dataset, write_dataset)
        self.rkom_bids = RKOMBidConfigurationsAPI(client, read_dataset, write_dataset)
        self.bechmarkings = BenchmarkingConfigurationsAPI(client, read_dataset, write_dataset)
        self.markets = MarketConfigurationsAPI(client, read_dataset, write_dataset)
        self.rkom_bid_combinations = RKOMBidCombinationConfiguration(client, read_dataset, write_dataset)


class PowerOpsClient:
    def __init__(
        self,
        read_dataset: str,
        write_dataset: str,
        cogshop_version: str,
        config: ClientConfig,
    ):
        self._read_dataset = read_dataset
        self._write_dataset = write_dataset
        self._cogshop_version = cogshop_version

        self.cdf = CogniteClient(config=config) if config else get_cognite_client()

        self.configurations = ConfigurationClient()

        self.shop = ShopAPI(po_client=self)

        self.configurations = ConfigurationsClient(read_dataset, write_dataset, self.cdf)
        self.generators = GeneratorsAPI(self.cdf, read_dataset, write_dataset)
        self.plants = PlantAPI(self.cdf, read_dataset, write_dataset)
        self.price_areas = PriceAreasAPI(self.cdf, read_dataset, write_dataset)
        self.reservoirs = ReservoirsAPI(self.cdf, read_dataset, write_dataset)
        self.watercourses = WatercourseAPI(self.cdf, read_dataset, write_dataset)

    @classmethod
    def from_settings(cls):
        return cls(
            read_dataset=settings.powerops.read_dataset,
            write_dataset=settings.powerops.write_dataset,
            cogshop_version=settings.powerops.cogshop_version,
            config=get_client_config(),
        )

    @cached_property
    def read_dataset_id(self) -> int:
        return retrieve_dataset(self.cdf, external_id=self._read_dataset).id

    @cached_property
    def write_dataset_id(self) -> int:
        return retrieve_dataset(self.cdf, external_id=self._write_dataset).id

    @property
    def cogshop_version(self) -> str:
        return self._cogshop_version



from typing import Any

from cognite.powerops.data_classes.time_series_mapping import TimeSeriesMapping, TimeSeriesMappingEntry
from cognite.powerops.data_classes.transformation import Transformation, TransformationType

ignored_attributes = [
    "gate.schedule_flag",
    "gate.schedule_m3s",
    "gate.schedule_percent",
    "generator.committed_flag",
    "generator.committed_in",
    "generator.maintenance_flag",
    "generator.production_schedule",
    "generator.production_schedule_flag",
    "market.buy_price",
    "market.sale_price",
    "plant.feeding_fee",
    "plant.maintenance_flag",
    "plant.production_schedule",
    "plant.production_schedule_flag",
    "reservoir.inflow",
    "reservoir.schedule",
    "reservoir.schedule_flag",
    "reservoir.min_vol_constr",
    "reservoir.max_vol_constr",
    "reservoir.tactical_cost_max",
    "reservoir.tactical_cost_min",
    "reservoir.tacical_limit_max",
    "reservoir.tactical_limit_min",
]


def is_constant_valued_dict(attribute_value: Any) -> bool:
    """Check if all dictionary values are the same"""
    return (
        isinstance(attribute_value, dict)
        and all(isinstance(value, (float, int)) for value in attribute_value.values())
        and len(set(attribute_value.values())) == 1
    )


def get_static_mapping(shop_model: dict) -> TimeSeriesMapping:
    # sourcery skip: for-append-to-extend, merge-nested-ifs
    static_mapping = TimeSeriesMapping()
    # e.g. reservoir: [reservoir_A, reservoir_B]
    for object_type, objects in shop_model.items():
        # e.g. reservoir_A: [maintenance_flag, max_vol, ...]
        for object_name, object_attributes in objects.items():
            # e.g. max_vol: 42 or maintenance_flag: {<date>: 1, <other_date>: 0}
            for attribute_name, attribute_value in object_attributes.items():
                if f"{object_type}.{attribute_name}" in ignored_attributes:
                    continue

                if is_constant_valued_dict(attribute_value):
                    static_value = {0: list(attribute_value.values())[0]}  # TODO: clean up
                    static_mapping.append(
                        TimeSeriesMappingEntry(
                            object_type=object_type,
                            object_name=str(object_name),
                            attribute_name=attribute_name,
                            transformations=[
                                Transformation(transformation=TransformationType.STATIC, kwargs=static_value)
                            ],
                        )
                    )

    return static_mapping



from __future__ import annotations

from pathlib import Path
from typing import Annotated

import typer

from cognite.powerops.bootstrap import (
    _create_cdf_resources,
    _load_config,
    _preview_resources_diff,
    _transform,
    validate_config,
)
from cognite.powerops.logger import configure_debug_logging

app = typer.Typer()


@app.command(
    "plan",
    help="Preview the changes from the configuration files that `powerops apply` would make to the data model in CDF",
)
def plan(
    path: Annotated[Path, typer.Argument(help="Path to configuration files")],
    market: Annotated[str, typer.Argument(help="Selected power market")],
):
    typer.echo(f"Running plan on configuration files located in {path}")

    # Step 1 - configure and validate config
    config = _load_config(path)
    config = validate_config(config)
    configure_debug_logging(config.constants.debug_level)

    # Step 2 - transform from config to CDF resources and preview diffs
    # 2.a - transform
    bootstrap_resources = _transform(
        config,
        path,
        market,
    )

    # 2.b - preview diff
    _preview_resources_diff(
        bootstrap_resources,
        config.constants.data_set_external_id,
    )


@app.command("apply", help="Apply the changes from the configuration files to the data model in CDF")
def apply(
    path: Annotated[Path, typer.Argument(help="Path to configuration files")],
    market: Annotated[str, typer.Argument(help="Selected power market")],
):
    typer.echo(f"Running apply on configuration files located in {path}")

    # Step 1 - configure and validate config
    config = _load_config(path)
    config = validate_config(config)
    configure_debug_logging(config.constants.debug_level)

    # Step 2 - transform from config to CDF resources and preview diffs
    bootstrap_resources = _transform(
        config,
        path,
        market,
    )

    # Step 3 - write bootstrap resources from diffs to CDF
    _create_cdf_resources(
        bootstrap_resources,
        config.constants.data_set_external_id,
        config.constants.overwrite_data,
        config.constants.skip_dm,
    )


def main():
    app()



import datetime
from uuid import uuid4

from cognite.client.data_classes import Event


def create_bootstrap_finished_event() -> Event:
    """Creating a POWEROPS_BOOTSTRAP_FINISHED Event in CDF to signal that bootstrap scripts have been ran"""
    current_time = int(datetime.datetime.utcnow().timestamp() * 1000)  # in milliseconds
    event = Event(
        start_time=current_time,
        end_time=current_time,
        external_id=f"POWEROPS_BOOTSTRAP_FINISHED_{str(uuid4())}",
        type="POWEROPS_BOOTSTRAP_FINISHED",
        subtype=None,
        source="PowerOps bootstrap",
        description="Manual run of bootstrap scripts finsihed",
    )
    print(f"Created status event '{event.external_id}'")

    return event



from __future__ import annotations

import logging
import random
from typing import TYPE_CHECKING

from cognite.powerops.client.api.shop_result_files_api import ShopFilesAPI
from cognite.powerops.client.api.shop_results_api import ShopRunResultsAPI
from cognite.powerops.client.api.shop_run_api import ShopRunsAPI

if TYPE_CHECKING:
    from cognite.powerops import PowerOpsClient

logger = logging.getLogger(__name__)


class ShopModel:
    def __init__(self) -> None:
        self.model_id = random.randint(1000, 9999)

    def render_yaml(self) -> str:
        return "sintef_shop_model_yaml_representation"

    def update(self):
        raise NotImplementedError


class ShopModelsAPI:
    def __init__(self, po_client: PowerOpsClient):
        self._po_client = po_client

    def list(self) -> list[ShopModel]:
        return [ShopModel()]

    def retrieve(self, model_id):
        m = ShopModel()
        m.model_id = model_id
        return m


class ShopAPI:
    def __init__(self, po_client: PowerOpsClient):
        self._po_client = po_client
        self.models = ShopModelsAPI(po_client)
        self.runs = ShopRunsAPI(po_client)
        self.results = ShopRunResultsAPI(po_client)
        self.files = ShopFilesAPI(po_client)



from typing import Optional

import pydantic

from cognite.powerops.settings.loader import file_settings

__all__ = ["Settings"]


class CogniteSettings(pydantic.BaseModel):
    project: Optional[str]
    cdf_cluster: Optional[str]
    tenant_id: Optional[str]
    client_id: Optional[str]
    client_secret: Optional[str]
    space: Optional[str]
    data_model: Optional[str]
    schema_version: Optional[str]


class PoweropsRunSettings(pydantic.BaseModel):
    read_dataset: Optional[str]
    write_dataset: Optional[str]
    cogshop_version: Optional[str]


class Settings(pydantic.BaseSettings):
    cognite: CogniteSettings = {}
    powerops: PoweropsRunSettings = {}

    class Config:
        env_prefix = "SETTINGS__"
        env_nested_delimiter = "__"

        @classmethod
        def customise_sources(cls, init_settings, env_settings, file_secret_settings):
            """Add `file_settings` to sources (loads settings.toml and .secrets.toml)."""
            return init_settings, env_settings, file_settings, file_secret_settings



import logging
import os
from typing import Any

try:
    import tomllib
except ModuleNotFoundError:
    import tomli as tomllib  # py < 3.11

from pydantic import BaseSettings

from cognite.powerops.utils.common import merge

logger = logging.getLogger(__name__)


def file_settings(_settings: BaseSettings) -> dict[str, Any]:
    settings_files = filter(None, os.environ.get("SETTINGS_FILES", "settings.toml;.secrets.toml").split(";"))
    collected_data = {}
    for file_path in settings_files:
        try:
            data = _load_file(file_path)
        except FileNotFoundError:
            pass
        else:
            logger.debug(f"Loaded settings from '{file_path}'.")
            collected_data = merge(collected_data, data)
    return collected_data


def _load_file(file_path: str) -> dict:
    with open(file_path, "rb") as toml_file:
        return tomllib.load(toml_file)



from typing import Optional

from cognite.client import CogniteClient
from cognite.client.exceptions import CogniteReadTimeout
from pydantic import BaseModel, Field

from cognite.powerops.utils.retry import retry


class Transformation(BaseModel):
    method: str
    arguments: Optional[str]


class TransformationItems(BaseModel):
    items: list[Transformation]


class Commands(BaseModel):
    commands: list[str]


class FileRef(BaseModel):
    file_external_id: str = Field(alias="fileExternalId")
    type: Optional[str]


class FileRefItems(BaseModel):
    items: list[FileRef]


class Mapping(BaseModel):
    path: str
    timeseries_external_id: Optional[str] = Field(alias="timeseriesExternalId")
    transformations: Optional[TransformationItems]
    retrieve: Optional[str]
    aggregation: Optional[str]


class MappingItems(BaseModel):
    items: list[Mapping]


class ModelTemplate(BaseModel):
    watercourse: str
    model: FileRef
    base_mappings: MappingItems = Field(alias="baseMappings")


class Scenario(BaseModel):
    name: str
    model_template: ModelTemplate = Field(alias="modelTemplate")
    commands: Commands
    mappings_override: Optional[MappingItems] = Field(alias="mappingsOverride")
    extra_files: Optional[FileRefItems] = Field(alias="extraFiles")


class Case(BaseModel):
    start_time: str = Field(alias="startTime")  # datetime.datetime
    end_time: str = Field(alias="endTime")  # datetime.datetime
    scenario: Scenario


def render_query(space: str, case_external_id: str) -> str:
    query_template = """
    query MyQuery {{
    getCaseById(instance: {{space: "{space}", externalId: "{external_id}"}}) {{
        items {{
        startTime
        endTime
        scenario {{
            name
            modelTemplate {{
            watercourse
            model {{
                fileExternalId
            }}
            baseMappings (first: 500) {{
                items {{
                    path
                    timeseriesExternalId
                    retrieve
                    aggregation
                    transformations {{
                        items {{
                            method
                            arguments
                        }}
                    }}
                }}
            }}
            }}
            commands {{
                commands
            }}
        }}
        }}
    }}
    }}
    """
    return query_template.format(space=space, external_id=case_external_id)


def parse_response(response) -> list[Case]:
    data: dict = response.json()["data"]
    query_key = tuple(data.keys())[0]
    items = data[query_key]["items"]
    return [Case.parse_obj(item) for item in items]


@retry(CogniteReadTimeout, tries=5, delay=1, backoff=2, jitter=(0.1, 0.3))
def query_fdm(
    client: CogniteClient,
    space: str,
    case_external_id: str,
    model_extenal_id: Optional[str] = None,
    model_version: Optional[int] = None,
):
    project = client.iam.token.inspect().projects[0].url_name
    model = space if model_extenal_id is None else model_extenal_id
    version = str(model_version) if model_version is not None else "1"

    url = f"/api/v1/projects/{project}/userapis/spaces/{space}/datamodels/{model}/versions/{version}/graphql"
    return client.post(url, json={"query": render_query(space, case_external_id)})


def get_case(
    client: CogniteClient,
    space: str,
    case_external_id: str,
    model_extenal_id: Optional[str] = None,
    model_version: Optional[int] = None,
) -> Case:
    response = query_fdm(
        client=client,
        space=space,  # type: ignore[arg-type]
        case_external_id=case_external_id,  # type: ignore[arg-type]
        model_extenal_id=model_extenal_id,  # type: ignore[arg-type]
        model_version=model_version,  # type: ignore[arg-type]
    )

    return parse_response(response)[0]



from __future__ import annotations

import json
from collections import UserList
from typing import Any, Collection, Optional, Type, TypeVar

import pandas as pd
from cognite.client.utils._identifier import IdentifierSequence
from cognite.client.utils._pandas_helpers import convert_nullable_int_cols, notebook_display_with_fallback
from cognite.client.utils._time import convert_time_attributes_to_datetime
from pydantic import BaseModel

T_AssetResource = TypeVar("T_AssetResource", bound=BaseModel)


class AssetResourceList(UserList):
    _RESOURCE: Type[BaseModel]

    def __init__(self, resources: Collection[Any]):
        for resource in resources:
            if not isinstance(resource, self._RESOURCE):
                raise TypeError(
                    f"All resources for class '{self.__class__.__name__}' must be of type "
                    f"'{self._RESOURCE.__name__}', not '{type(resource)}'."
                )
        super().__init__(resources)
        self._id_to_item, self._external_id_to_item = {}, {}
        if self.data:
            if hasattr(self.data[0], "external_id"):
                self._external_id_to_item = {
                    item.external_id: item for item in self.data if item.external_id is not None
                }
            if hasattr(self.data[0], "id"):
                self._id_to_item = {item.id: item for item in self.data if item.id is not None}

    def __getitem__(self, item: Any) -> Any:
        value = super().__getitem__(item)
        if isinstance(item, slice):
            return self.__class__(value)
        return value

    def __str__(self) -> str:
        item = convert_time_attributes_to_datetime(self.dump())
        return json.dumps(item, indent=4)

    def extend(self, other: Collection[Any]) -> None:
        other_res_list = type(self)(other)  # See if we can accept the types
        if set(self._id_to_item).isdisjoint(other_res_list._id_to_item):
            super().extend(other)
            self._external_id_to_item.update(other_res_list._external_id_to_item)
            self._id_to_item.update(other_res_list._id_to_item)
        else:
            raise ValueError("Unable to extend as this would introduce duplicates")

    def dump(self) -> list[dict[str, Any]]:
        """Dump the instance into a json serializable Python data type.

        Args:
            camel_case (bool): Use camelCase for attribute names. Defaults to False.

        Returns:
            List[Dict[str, Any]]: A list of dicts representing the instance.
        """
        return [resource.dict(exclude_unset=True) for resource in self.data]

    def get(self, id: int = None, external_id: str = None) -> Optional[T_AssetResource]:
        """Get an item from this list by id or exernal_id.

        Args:
            id (int): The id of the item to get.
            external_id (str): The external_id of the item to get.

        Returns:
            Optional[CogniteResource]: The requested item
        """
        IdentifierSequence.load(id, external_id).assert_singleton()
        if id:
            return self._id_to_item.get(id)
        return self._external_id_to_item.get(external_id)

    def to_pandas(self, camel_case: bool = False) -> pd.DataFrame:
        """Convert the instance into a pandas DataFrame.

        Returns:
            pandas.DataFrame: The dataframe.
        """
        df = pd.DataFrame(self.dump())
        return convert_nullable_int_cols(df, camel_case)

    def _repr_html_(self) -> str:
        return notebook_display_with_fallback(self)


T_AssetResourceList = TypeVar("T_AssetResourceList", bound=AssetResourceList)



from typing import Optional

from cognite.client.data_classes import Asset, Label, Relationship
from pydantic import BaseModel

from cognite.powerops.config import GeneratorTimeSeriesMapping
from cognite.powerops.data_classes.cdf_resource_collection import BootstrapResourceCollection
from cognite.powerops.utils.labels import AssetLabels
from cognite.powerops.utils.labels import RelationshipLabels as rl
from cognite.powerops.utils.relationship_types import asset_to_time_series

ExternalId = str
GeneratorName = str


class Generator(BaseModel):
    name: GeneratorName
    penstock: str
    startcost: float
    p_min: float

    start_stop_cost_time_series: Optional[ExternalId] = None  # external ID of time series with values in m

    @property
    def external_id(self) -> ExternalId:
        return f"generator_{self.name}"

    @classmethod
    def add_time_series_mapping(
        cls,
        generator_time_series_mappings: list[GeneratorTimeSeriesMapping],
        generators: dict[GeneratorName, "Generator"],
    ):
        for mapping in generator_time_series_mappings:
            generator_name = mapping.generator_name
            # check if the generator is in the given watercourse (defined by the plants dict)
            if generator_name not in generators:
                continue

            generators[generator_name].start_stop_cost_time_series = mapping.start_stop_cost

    def to_bootstrap_resources(self) -> BootstrapResourceCollection:
        asset = self.asset()
        relationships = self.relationships()
        return BootstrapResourceCollection(
            assets={asset.external_id: asset}, relationships={rel.external_id: rel for rel in relationships}
        )

    def asset(self) -> Asset:
        return Asset(
            external_id=f"generator_{self.name}",
            name=self.name,
            parent_external_id="generators",
            labels=[Label(AssetLabels.GENERATOR)],
            metadata={
                "penstock": self.penstock,
                "startcost": self.startcost,
                "p_min": self.p_min,
            },
        )

    def relationships(self) -> list[Relationship]:
        time_series_to_append_if_not_none = {
            self.start_stop_cost_time_series: rl.START_STOP_COST_TIME_SERIES,
        }
        return [
            asset_to_time_series(self.external_id, time_series, label)
            for time_series, label in time_series_to_append_if_not_none.items()
            if time_series
        ]



from __future__ import annotations

from pathlib import Path
from typing import Optional, TypedDict

import pandas as pd
from cognite.client.data_classes import Asset, Relationship, Sequence

from cognite.powerops.config import GeneratorTimeSeriesMapping, PlantTimeSeriesMapping, WatercourseConfig
from cognite.powerops.data_classes.cdf_resource_collection import BootstrapResourceCollection, SequenceContent
from cognite.powerops.data_classes.generator import Generator
from cognite.powerops.data_classes.plant import Plant
from cognite.powerops.utils.asset_types import price_area_asset, reservoir_asset, watercourse_asset
from cognite.powerops.utils.common import replace_nordic_letters
from cognite.powerops.utils.relationship_types import (
    generator_to_generator_efficiency_curve,
    generator_to_turbine_efficiency_curve,
    price_area_to_dayahead_price,
    price_area_to_plant,
    price_area_to_watercourse,
    watercourse_to_plant,
    watercourse_to_production_obligation,
)
from cognite.powerops.utils.serializer import load_yaml


class ShopEfficiencyCurve(TypedDict):
    ref: str
    x: list[float]
    y: list[float]


def generate_resources_and_data(
    watercourse_configs: list[WatercourseConfig],
    plant_time_series_mappings: Optional[list[PlantTimeSeriesMapping]],
    generator_time_series_mappings: Optional[list[GeneratorTimeSeriesMapping]],
) -> BootstrapResourceCollection:
    """
    Create Assets for:
        - price_area,
        - watercourse
        - plant
        - reservoir
        - generator

    Create Relationships for:
        - price_area -> plant
        - price_area -> watercourse
        - watercourse -> plant
        - watercourse -> production obligation time series
        - plant -> reservoir
        - plant -> generator
        - generator -> generator efficiency
        - generator -> turbine efficiency

    Create sequences for
        - Generator efficiency
        - Turbine efficiency

    Create the sequence data for the
        - Generator efficiency
        - Turbine efficiency

    Parameters
    ----------
    watercourse_configs : List[WatercourseConfig]
        List of watercourse configs
    plant_time_series_mappings : List[PlantTimeSeriesMapping]
        List of plant time series mappings
    generator_time_series_mappings : List[GeneratorTimeSeriesMapping]
        List of generator time series mappings

    Returns
    -------
    CDFResourceCollection
        Collection of Assets, Sequences, Relationships and sequence data
    """

    resources = BootstrapResourceCollection()

    for watercourse_config in watercourse_configs:
        watercourse = watercourse_asset(
            name=watercourse_config.name,
            shop_penalty_limit=watercourse_config.shop_penalty_limit,
        )
        resources.add(watercourse)

        shop_case = load_yaml(Path(watercourse_config.yaml_raw_path), clean_data=True)

        reservoirs = shop_case["model"]["reservoir"]
        resources += create_reservoirs(reservoirs, watercourse_config)

        generators = shop_case["model"]["generator"]
        resources += add_generators_and_efficiency_curves(generators, generator_time_series_mappings)

        plants = Plant.from_shop_case(shop_case=shop_case)
        if plant_time_series_mappings:
            Plant.add_time_series_mapping(
                plant_time_series_mappings=plant_time_series_mappings,
                plants=plants,
            )
        # Add display name and ordering key.
        for plant in plants.values():
            plant.display_name = watercourse_config.plant_display_name(plant.name)
            plant.ordering_key = watercourse_config.plant_ordering_key(plant.name)

            resources += plant.to_bootstrap_resources()

        # Add relationships that are not covered by the Plant class
        for plant_name, plant_attributes in shop_case["model"]["plant"].items():
            # Using the first value of the prod_area time series as the prod_area (assuming it does not change
            prod_area = str(list(plant_attributes["prod_area"].values())[0])
            price_area_name = watercourse_config.market_to_price_area[prod_area]
            price_area_external_id = f"price_area_{price_area_name}"
            if price_area_external_id not in resources.assets:
                resources.add(price_area_asset(name=price_area_name))

            plant_external_id = f"plant_{plant_name}"
            resources.add(watercourse_to_plant(watercourse=watercourse, plant=plant_external_id))
            resources.add(price_area_to_plant(price_area=price_area_external_id, plant=plant_external_id))
            resources.add(price_area_to_watercourse(price_area=price_area_external_id, watercourse=watercourse))

        if watercourse_config.production_obligation_ts_ext_ids:
            for ts_ext_id in watercourse_config.production_obligation_ts_ext_ids:
                rel = watercourse_to_production_obligation(
                    watercourse=watercourse, production_obligation_ts_ext_id=ts_ext_id
                )
                resources.add(rel)

    return resources


def create_generator_efficiency_curve_sequence(
    efficiency_curve: ShopEfficiencyCurve,
    generator_asset: Asset,
) -> BootstrapResourceCollection:
    """
    Create the cognite resource for the generator efficiency curve (Sequence) and the connection
    (Relationship) between the generator and the efficiency curve.

    Parameters
    ----------
    efficiency_curve : dict[Literal["ref", "x", "y"], float | list[float]]
        Dictionary of the generator efficiency curve
    generator_asset : Asset
        The generator asset to which the efficiency curve belongs

    Returns
    -------
    BootstrapResourceCollection
        The new resources (incl. sequence data)
    """
    resources = BootstrapResourceCollection()
    x_col_name = "generator_power"
    y_col_name = "generator_efficiency"

    sequence = Sequence(
        external_id=f"{generator_asset.external_id}_generator_efficiency_curve",
        name=f"{generator_asset.name} generator efficiency curve",
        columns=[
            {"valueType": "DOUBLE", "externalId": x_col_name},
            {"valueType": "DOUBLE", "externalId": y_col_name},
        ],
    )
    resources.add(sequence)
    resources.add(
        SequenceContent(
            sequence_external_id=sequence.external_id,
            data=pd.DataFrame(
                {
                    x_col_name: efficiency_curve["x"],
                    y_col_name: efficiency_curve["y"],
                },
                dtype=float,
            ),
        )
    )

    relationship = generator_to_generator_efficiency_curve(
        generator=generator_asset, generator_efficiency_curve=sequence
    )

    resources.add(relationship)

    return resources


def create_turbine_efficiency_curve_sequence(
    efficiency_curves: list[ShopEfficiencyCurve],
    generator_asset: Asset,
) -> BootstrapResourceCollection:
    """
    Create the cognite resource for turbine efficiency curve (Sequence) and the connection
    (Relationship) between the generator and the efficiency curve.

    Parameters
    ----------
    efficiency_curves : list[dict[Literal["ref", "x", "y"], float | list[float]]]
        List of dictionaries of the turbine efficiency curves
    generator_asset : Asset
        The generator asset to which the efficiency curve belongs

    Returns
    -------
    tuple[ResourceDict, dict[str, pd.DataFrame]]
        The new resources and sequence data dictionary
    """
    resources = BootstrapResourceCollection()
    ref_col_name = "head"
    x_col_name = "flow"
    y_col_name = "turbine_efficiency"

    sequence_external_id = f"{generator_asset.external_id}_turbine_efficiency_curve"
    # Start by creating the sequence
    sequence = Sequence(
        external_id=sequence_external_id,
        name=f"{generator_asset.name} turbine efficiency curve",
        columns=[
            {"valueType": "DOUBLE", "externalId": ref_col_name},
            {"valueType": "DOUBLE", "externalId": x_col_name},
            {"valueType": "DOUBLE", "externalId": y_col_name},
        ],
    )
    resources.add(sequence)

    # Then create the sequence data
    df_list = [
        pd.DataFrame(
            {
                ref_col_name: [],
                x_col_name: [],
                y_col_name: [],
            }
        )
    ]

    for efficiency_curve in efficiency_curves:
        ref = efficiency_curve["ref"]
        x = efficiency_curve["x"]
        y = efficiency_curve["y"]
        temp_df = pd.DataFrame(
            {
                ref_col_name: [ref] * len(x),
                x_col_name: x,
                y_col_name: y,
            }
        )
        df_list.append(temp_df)
    df = pd.concat(df_list, ignore_index=True)

    resources.add(
        SequenceContent(
            sequence_external_id=sequence.external_id,
            data=df,
        )
    )

    # Finally create the relationship
    relationship = generator_to_turbine_efficiency_curve(generator=generator_asset, turbine_efficiency_curve=sequence)
    resources.add(relationship)

    return resources


def get_single_value(value_or_time_series: float | dict) -> float:
    """Get the single value from a time series or a value
    returns the value if value_or_time_series is a value, otherwise the first value in the time series

    Parameters
    ----------
    value_or_time_series : float | dict
        Either a simple numeric value, or a dictionary (time series, {datetime string: value})
    """
    if isinstance(value_or_time_series, dict):
        return next(iter(value_or_time_series.values()))
    return value_or_time_series


def add_generators_and_efficiency_curves(
    shop_generator_dict: dict,
    generator_time_series_mappings: Optional[list[GeneratorTimeSeriesMapping]],
) -> BootstrapResourceCollection:
    """Create the cognite resource for the generator (Asset) and the generator- and turbine efficiencies (sequences), as
        well as the connection (Relationship) between the generator and the efficiency curves.

    Parameters
    ----------
    shop_generator_dict : dict
        Dictionary of generators objects from the shop case file
    generator_time_series_mappings : GeneratorTimeSeriesMapping
        Mapping of generator time series

    Returns
    -------
    BootstrapResourceCollection
        Collection of assets, sequences, relationships and sequence data
    """
    generators: {str: Generator} = {}
    resources = BootstrapResourceCollection()
    for generator_name, generator_information in shop_generator_dict.items():
        generator_name = replace_nordic_letters(generator_name)

        generator = Generator(
            name=generator_name,
            penstock=str(generator_information.get("penstock", "1")),
            p_min=generator_information.get("p_min", 0.0),
            startcost=get_single_value(generator_information.get("startcost", 0.0)),
        )
        generators[generator_name] = generator
        generator_asset = generator.asset()

        resources += create_generator_efficiency_curve_sequence(
            generator_information["gen_eff_curve"],
            generator_asset,
        )

        resources += create_turbine_efficiency_curve_sequence(
            generator_information["turb_eff_curves"],
            generator_asset,
        )

    if generator_time_series_mappings:
        Generator.add_time_series_mapping(
            generator_time_series_mappings=generator_time_series_mappings, generators=generators
        )

    for generator in generators.values():
        resources += generator.to_bootstrap_resources()

    return resources


def create_reservoirs(
    reservoirs: dict,
    watercourse_config: WatercourseConfig,
) -> BootstrapResourceCollection:
    resources = BootstrapResourceCollection()
    for reservoir_name in reservoirs:
        reservoir_name = replace_nordic_letters(reservoir_name)

        reservoir = reservoir_asset(
            name=reservoir_name,
            display_name=watercourse_config.reservoir_display_name(reservoir_name),
            ordering_key=watercourse_config.reservoir_ordering_key(reservoir_name),
        )
        resources.add(reservoir)

    return resources


def generate_relationships_from_price_area_to_price(
    price_ts_ext_id_per_price_area: dict[str, str]
) -> list[Relationship]:
    return [
        price_area_to_dayahead_price(price_area_asset(price_area_name), ts_ext_id)
        for price_area_name, ts_ext_id in price_ts_ext_id_per_price_area.items()
    ]



from dataclasses import dataclass
from typing import List, Optional

from cognite.client.data_classes import Sequence
from pandas import DataFrame

from cognite.powerops.data_classes.cdf_resource_collection import BootstrapResourceCollection, SequenceContent


@dataclass
class ShopOutputDefinition:
    shop_object_type: str
    shop_attribute_name: str
    cdf_attribute_name: str
    unit: str
    is_step: bool = True


@dataclass
class ShopOutputConfig:
    watercourse: str
    shop_output_definitions: Optional[List[ShopOutputDefinition]] = None

    @property
    def default_output_definitions(self) -> List[ShopOutputDefinition]:
        return [
            ShopOutputDefinition("market", "sale_price", "price", "EUR/MWh"),
            ShopOutputDefinition("market", "sale", "sales", "MWh"),
            ShopOutputDefinition("plant", "production", "production", "MW"),
            ShopOutputDefinition("plant", "consumption", "consumption", "MW"),
            ShopOutputDefinition("reservoir", "water_value_global_result", "water_value", "EUR/Mm3"),
            ShopOutputDefinition("reservoir", "energy_conversion_factor", "energy_conversion_factor", "MWh/Mm3"),
        ]

    def to_dataframe(self):
        if not self.shop_output_definitions:
            self.shop_output_definitions = self.default_output_definitions
        df = DataFrame(self.shop_output_definitions)
        df["is_step"] = df["is_step"].map({True: "True", False: "False"})
        return df

    @classmethod
    def to_sequence(cls, watercourse: str) -> Sequence:
        sequence_external_id = f"SHOP_{watercourse.replace(' ', '_')}_output_definition"
        column_def = [
            {"valueType": "STRING", "externalId": "shop_object_type"},
            {"valueType": "STRING", "externalId": "shop_attribute_name"},
            {"valueType": "STRING", "externalId": "cdf_attribute_name"},
            {"valueType": "STRING", "externalId": "unit"},
            {"valueType": "STRING", "externalId": "is_step"},
        ]

        return Sequence(
            name=sequence_external_id.replace("_", " "),
            description="Defining which SHOP results to output to CDF (as time series)",
            external_id=sequence_external_id,
            columns=column_def,
            metadata={
                "shop:watercourse": watercourse,
                "shop:type": "output_definition",
            },
        )

    def to_bootstrap_resources(self) -> BootstrapResourceCollection:
        bootstrap_resources = BootstrapResourceCollection()
        sequence = ShopOutputConfig.to_sequence(watercourse=self.watercourse)

        bootstrap_resources.add(sequence)
        bootstrap_resources.add(SequenceContent(sequence_external_id=sequence.external_id, data=self.to_dataframe()))
        return bootstrap_resources



from __future__ import annotations

import json
from pathlib import Path
from typing import List, NamedTuple, Optional, Tuple

from cognite.client.data_classes import Asset, Label, Relationship
from pydantic import BaseModel, validator

from cognite.powerops.config import PlantTimeSeriesMapping
from cognite.powerops.data_classes.cdf_resource_collection import BootstrapResourceCollection
from cognite.powerops.data_classes.time_series_mapping import TimeSeriesMapping
from cognite.powerops.utils.common import print_warning
from cognite.powerops.utils.labels import AssetLabels as al
from cognite.powerops.utils.labels import RelationshipLabels as rl
from cognite.powerops.utils.relationship_types import asset_to_time_series, plant_to_generator, plant_to_inlet_reservoir
from cognite.powerops.utils.serializer import load_yaml

p_min_fallback = 0
p_max_fallback = 1e20

head_loss_factor_fallback = 0.0

ExternalId = str
PlantName = str


class Plant(BaseModel):
    name: PlantName
    external_id: ExternalId
    outlet_level: float  # meters above sea level
    p_min: float = p_min_fallback
    p_max: float = p_max_fallback  # arbitrary max power output if not specified
    head_loss_factor: float = 0.0
    penstock_head_loss_factors: Optional[dict[str, float]] = None
    display_name: Optional[str] = None
    ordering_key: Optional[float] = None

    generator_ext_ids: list[ExternalId] = []  # external IDs of generator assets
    inlet_reservoir_ext_id: Optional[ExternalId] = None  # external ID of reservoir asset

    # TODO: Consider splitting this into a separate class/subclass
    inlet_level_time_series: Optional[ExternalId] = None  # external ID of time series with values in m.a.s.l.
    outlet_level_time_series: Optional[ExternalId] = None  # external ID of time series with values in m.a.s.l.
    water_value_time_series: Optional[ExternalId] = None  # external ID of time series with values in €/MWh
    feeding_fee_time_series: Optional[ExternalId] = None  # external ID of time series with values in percent
    p_min_time_series: Optional[ExternalId] = None  # external ID of time series with values in MW
    p_max_time_series: Optional[ExternalId] = None  # external ID of time series with values in MW
    head_direct_time_series: Optional[ExternalId] = None  # external ID of time series with values in m

    @validator("penstock_head_loss_factors", pre=True)
    def parse_dict(cls, value):
        if isinstance(value, str):
            value = json.loads(value)
        return value

    def asset(self) -> Asset:
        """Returns the Asset representation of this Plant (without relationships)"""
        return Asset(
            name=self.name,
            external_id=self.external_id,
            parent_external_id="plants",
            metadata={
                "outlet_level": self.outlet_level,
                "p_min": self.p_min,
                "p_max": self.p_max,
                "head_loss_factor": self.head_loss_factor,
                "penstock_head_loss_factors": json.dumps(self.penstock_head_loss_factors),
                "display_name": self.display_name or self.name,
                "ordering": str(self.ordering_key or 999),
            },
            labels=[Label(external_id=al.PLANT)],
        )

    def relationships(self) -> list[Relationship]:
        time_series_to_append_if_not_none = {
            self.inlet_level_time_series: rl.INLET_LEVEL_TIME_SERIES,
            self.outlet_level_time_series: rl.OUTLET_LEVEL_TIME_SERIES,
            self.water_value_time_series: rl.WATER_VALUE_TIME_SERIES,
            self.feeding_fee_time_series: rl.FEEDING_FEE_TIME_SERIES,
            self.p_min_time_series: rl.P_MIN_TIME_SERIES,
            self.p_max_time_series: rl.P_MAX_TIME_SERIES,
            self.head_direct_time_series: rl.HEAD_DIRECT_TIME_SERIES,
        }
        relationships: list[Relationship] = [
            asset_to_time_series(self.external_id, time_series, label)
            for time_series, label in time_series_to_append_if_not_none.items()
            if time_series
        ]

        relationships.extend(
            plant_to_generator(plant=self.external_id, generator=generator) for generator in self.generator_ext_ids
        )
        if self.inlet_reservoir_ext_id:
            relationships.append(
                plant_to_inlet_reservoir(plant=self.external_id, reservoir=self.inlet_reservoir_ext_id)
            )

        return relationships

    def to_bootstrap_resources(self) -> BootstrapResourceCollection:
        asset = self.asset()
        relationships = self.relationships()
        return BootstrapResourceCollection(
            assets={asset.external_id: asset}, relationships={rel.external_id: rel for rel in relationships}
        )

    @classmethod
    def from_cdf_resources(cls, asset: Asset, relationships: list[Relationship], **kwargs) -> "Plant":
        """Initialise a Plant from CDF Asset and Relationships

        Args: asset (Asset): The plant Asset relationships (list[Relationship]): Relationships to related resources (
        will be mapped to attributes based on labels) **kwargs: Any other attributes that are not part of the Asset
        """
        # Initialise plant from Asset
        plant = cls(
            name=asset.name,
            external_id=asset.external_id,
            outlet_level=asset.metadata["outlet_level"],
            p_min=asset.metadata["p_min"],
            p_max=asset.metadata["p_max"],
            head_loss_factor=asset.metadata["head_loss_factor"],
            penstock_head_loss_factors=json.loads(asset.metadata.get("penstock_head_loss_factors")),
            **kwargs,  # kwargs to set any other attributes that are not part of the Asset
        )

        # Find time series based on relationships
        time_series_ext_ids_and_labels = [
            (rel.target_external_id, [label.external_id for label in rel.labels])
            for rel in relationships
            if rel.target_type == "TIMESERIES" and rel.source_external_id == plant.external_id
        ]
        for ts_ext_id, labels in time_series_ext_ids_and_labels:
            if rl.INLET_LEVEL_TIME_SERIES in labels:
                plant.inlet_level_time_series = ts_ext_id
            elif rl.OUTLET_LEVEL_TIME_SERIES in labels:
                plant.outlet_level_time_series = ts_ext_id
            elif rl.WATER_VALUE_TIME_SERIES in labels:
                plant.water_value_time_series = ts_ext_id
            elif rl.FEEDING_FEE_TIME_SERIES in labels:
                plant.feeding_fee_time_series = ts_ext_id
            elif rl.P_MIN_TIME_SERIES in labels:
                plant.p_min_time_series = ts_ext_id
            elif rl.P_MAX_TIME_SERIES in labels:
                plant.p_max_time_series = ts_ext_id
            elif rl.HEAD_DIRECT_TIME_SERIES in labels:
                plant.head_direct_time_series = ts_ext_id

        # Find generators based on relationships
        plant.generator_ext_ids = [
            rel.target_external_id for rel in relationships if label_in_labels(rl.GENERATOR, rel.labels)
        ]

        # Find inlet reservoir based on relationships
        for rel in relationships:
            if (
                rel.target_type == "ASSET"
                and label_in_labels(rl.INLET_RESERVOIR, rel.labels)
                and rel.source_external_id == plant.external_id
            ):
                plant.inlet_reservoir_ext_id = rel.target_external_id
                break

        return plant

    @classmethod
    def from_shop_case(cls, shop_case: dict, kwargs_per_plant: dict | None = None) -> dict[str, "Plant"]:
        """Returns a dict of Plant objects from a shop case dict
        Plants are only populated with the data that is in the shop case
        """
        if not kwargs_per_plant:
            kwargs_per_plant = {}
        plants = {
            name: cls(
                name=name,
                external_id=f"plant_{name}",
                outlet_level=attributes.get("outlet_line", 0),
                p_min=attributes.get("p_min", p_min_fallback),
                p_max=attributes.get("p_max", p_max_fallback),
                head_loss_factor=attributes.get("main_loss", [head_loss_factor_fallback])[
                    0
                ],  # For some reason, SHOP defines this as a list, but we only need the first (and only) value
                penstock_head_loss_factors={
                    str(penstock_number): loss_factor
                    for penstock_number, loss_factor in enumerate(
                        attributes.get("penstock_loss", [head_loss_factor_fallback]), start=1
                    )
                },
                **kwargs_per_plant.get(name, {}),
            )
            for name, attributes in shop_case["model"]["plant"].items()
        }
        # Add inlet reservoirs and generators from shop_case["connections"]
        all_connections = shop_case["connections"]
        reservoirs = shop_case["model"]["reservoir"]
        generators = shop_case["model"]["generator"]
        for plant_name, plant in plants.items():
            reservoir_name = plant_to_inlet_reservoir_breadth_first_search(plant_name, all_connections, reservoirs)
            plant.inlet_reservoir_ext_id = f"reservoir_{reservoir_name}"
            plant.generator_ext_ids = [
                f"generator_{gen_name}" for gen_name in generators_for_plant(plant_name, all_connections, generators)
            ]

        return plants

    @classmethod
    def add_time_series_mapping(
        cls,
        plant_time_series_mappings: list[PlantTimeSeriesMapping],
        plants: dict[PlantName, "Plant"],
    ) -> None:
        for mapping in plant_time_series_mappings:
            plant_name = mapping.plant_name
            # check if the plant is in the given watercourse (defined by the plants dict)
            if plant_name not in plants:
                continue

            plants[plant_name].water_value_time_series = mapping.water_value
            plants[plant_name].inlet_level_time_series = mapping.inlet_reservoir_level
            plants[plant_name].outlet_level_time_series = mapping.outlet_reservoir_level
            plants[plant_name].feeding_fee_time_series = mapping.feeding_fee
            plants[plant_name].p_min_time_series = mapping.p_min
            plants[plant_name].p_max_time_series = mapping.p_max
            plants[plant_name].head_direct_time_series = mapping.head_direct


def label_in_labels(label_external_id: str, labels: list[Label]) -> bool:
    """Check if a label with a given external id is in a list of labels."""
    return label_external_id in [label.external_id for label in labels]


def plant_to_inlet_reservoir_breadth_first_search(
    plant_name: str,
    all_connections: list[dict],
    reservoirs: dict,
) -> Optional[str]:
    """Search for a reservoir connected to a plant, starting from the plant and searching breadth first.

    Parameters
    ----------
    plant_name : str
        The plant we want to find a connection from
    all_connections : list[dict]
        All connections in the model.
    reservoirs : dict
        All reservoirs in the model. Keys are reservoir names.

    Returns
    -------
    Optional[str]
        The name of the reservoir connected to the plant, or None if no reservoir was found.
    """
    queue = []
    for connection in all_connections:
        if (
            connection["to"] == plant_name and connection.get("to_type", "plant") == "plant"
        ):  # if to_type is specified, it must be "plant"
            queue.append(connection)
            break
    visited = []
    while queue:
        connection = queue.pop(0)
        if connection not in visited:
            # Check if the given connection is from a reservoir
            # If we have "from_type" we can check directly if the object is a reservoir
            try:
                if connection["from_type"] == "reservoir":
                    return connection["from"]
            # If we don't have "from_type" we have to check if the name of the object is in the
            # list of reservoirs
            except KeyError:
                if connection["from"] in reservoirs:
                    return connection["from"]

            visited.append(connection)
            for candidate_connection in all_connections:
                # if the candidate connection is extension from the current connection, traverse it
                if candidate_connection["to"] == connection["from"]:
                    queue.append(candidate_connection)
    return None


def generators_for_plant(plant_name: str, all_connections: list[dict], generators) -> list[str]:
    """Returns a list of generators connected to a plant.

    Parameters
    ----------
    plant_name : str
        The name of the plant
    all_connections : list[dict]
        All connections in the model
    generators : list[str]
        Used to determine if the target of a connection is a generator (if to_type is not specified)

    Returns
    -------
    list[str]
        A list of generator names connected to the plant
    """
    return [
        connection["from"]
        for connection in all_connections
        if connection["to"] == plant_name
        and connection.get("to_type", "plant") == "plant"
        and (connection.get("from_type") == "generator" or connection["from"] in generators)
    ] + [  # Connections between plant and generator can be defined in either direction (it seems)
        connection["to"]
        for connection in all_connections
        if connection["from"] == plant_name
        and connection.get("from_type", "plant") == "plant"
        and (connection.get("to_type") == "generator" or connection["to"] in generators)
    ]


class Connection(NamedTuple):
    to_name: str
    from_name: str
    from_type: str
    to_type: str


def mip_plants(reserve_group: str, shop_case_yaml: str) -> List[str]:
    shop_case = load_yaml(Path(shop_case_yaml), clean_data=True)
    raw_connections = shop_case["connections"]
    connections = [
        Connection(
            to_name=str(connection["to"]),
            from_name=str(connection["from"]),
            to_type=connection["to_type"],
            from_type=connection["from_type"],
        )
        for connection in raw_connections
    ]
    # Find generators in the reserve groups
    reserve_group_to_generator = [
        connection.to_name
        for connection in connections
        if connection.to_type == "generator"
        and connection.from_type == "reserve_group"
        and connection.from_name == reserve_group
    ]
    generator_to_reserve_group = [
        connection.from_name
        for connection in connections
        if connection.from_type == "generator"
        and connection.to_type == "reserve_group"
        and connection.to_name == reserve_group
    ]
    generators_in_reserve_group = [*reserve_group_to_generator, *generator_to_reserve_group]

    # Find plants that the generators belong to
    generator_to_plant = [
        connection.to_name
        for connection in connections
        if connection.to_type == "plant"
        and connection.from_type == "generator"
        and connection.from_name in generators_in_reserve_group
    ]
    plant_to_generator = [
        connection.from_name
        for connection in connections
        if connection.from_type == "plant"
        and connection.to_type == "generator"
        and connection.to_name in generators_in_reserve_group
    ]
    return list({*generator_to_plant, *plant_to_generator})


def _match_plants_and_mip_flag(plants: List[str], mapping: TimeSeriesMapping) -> List[Tuple[str, Optional[str]]]:
    res: List[Tuple[str, Optional[str]]] = []
    for plant in plants:
        found = False
        for row in mapping:
            if row.object_name == plant and row.object_type == "plant" and row.attribute_name == "mip_flag":
                res.append((plant, row.time_series_external_id))
                print(f"\033[92m{plant}.mip_flag = {row.time_series_external_id}\033[0m")
                found = True
                break
        if not found:
            res.append((plant, None))
            print_warning(f"Did not find the `mip_flag` time series for {plant}!")
    print("\033[92m FINISHED \033[0m")
    return res



from __future__ import annotations

import logging
import tempfile
from typing import List, Optional, TypedDict

import yaml

logger = logging.getLogger(__name__)


class FileRefT(TypedDict):
    file: str
    encoding: str


class Case:
    r"""
    Wrapper around YAML file for SHOP, describing a case.

    Main features:
     * Values can be retrieved or set / changed like with a dict.
     * Can hold references to additional files ("extra files").
        * These are just paths to files, their content is not accessed here.
     * Loading from string or file (with `Case.from_yaml_file(path_to_yaml)`).
     * Saving to yaml (with `case.save_yaml(path_to_file)`).

    Examples:
      * load a case from string
          >>> case = Case('''
          ... foo:
          ...   bar1: 11
          ...   bar2: 22
          ... ''')
          >>> case.data
          {'foo': {'bar1': 11, 'bar2': 22}}

      * edit a value and show updated data
          >>> case.data["foo"]["bar2"] = 202
          >>> case.yaml
          'foo:\n  bar1: 11\n  bar2: 202\n'

      * load from and save to a file:
          case = Case.from_yaml_file("path/to/my_case.yaml")
          case[...] = ...  # edit data
          case.save_yaml("path/to/same_or_different.yaml")
    """

    def __init__(self, data: str = "") -> None:
        self._cut_file: Optional[FileRefT] = None
        self._extra_files: list[FileRefT] = []
        self._mapping_files: list[FileRefT] = []

        yaml_docs = list(yaml.safe_load_all(data))
        if yaml_docs:
            self.data = yaml_docs[0]
            self._handle_additional_yaml_documents(yaml_docs[1:])

    def _handle_additional_yaml_documents(self, extra_yaml_docs: List[str]) -> None:
        """
        If `Case.__init__` gets a yaml string which has multiple documents (separated by "---"),
        only the first document is parsed and set to `self.data`. Any subsequent documents are stored
        as "extra files". They are not part of `self.data`, but are not lost either.
        """
        if extra_yaml_docs:
            logger.warning(
                f"Case file contains {len(extra_yaml_docs) + 1} YAML documents. Only the first document is parsed,"
                f' additional documents will be passed to SHOP verbatim as "extra files".',
            )
        for yaml_doc in extra_yaml_docs:
            tmp_file = tempfile.NamedTemporaryFile(
                mode="w",
                delete=False,
                prefix="powerops-sdk-tmp-",
                suffix=".yaml",
            )
            tmp_file.write(yaml.dump(yaml_doc))
            tmp_file.close()
            self.add_extra_file(tmp_file.name)

    @classmethod
    def from_yaml_file(cls, yaml_path: str, encoding: str = "utf-8") -> Case:
        logger.info(f"loading case file: {yaml_path}")
        with open(yaml_path, "r", encoding=encoding) as yaml_file:
            return cls(yaml_file.read())

    def add_extra_file(self, file_path: str, encoding: str = "utf-8") -> None:
        logger.info(f"adding extra file: '{file_path}'")
        self._extra_files.append({"file": file_path, "encoding": encoding})

    def add_mapping_file(self, file_path: str, encoding: str = "utf-8") -> None:
        logger.info(f"adding mapping file: '{file_path}'")
        self._mapping_files.append({"file": file_path, "encoding": encoding})

    def add_cut_file(self, file_path: str, encoding: str = "utf-8") -> None:
        logger.info(f"adding cut file: '{file_path}'")
        self._cut_file = {"file": file_path, "encoding": encoding}

    @property
    def extra_files(self) -> list[FileRefT]:
        return self._extra_files

    @property
    def mapping_files(self) -> list[FileRefT]:
        return self._mapping_files

    @property
    def cut_file(self) -> FileRefT:
        return self._cut_file

    @property
    def yaml(self) -> str:
        return yaml.dump(self.data, sort_keys=False)

    def save_yaml(self, path: str, encoding: str = "utf-8") -> None:
        logger.info(f"Saving case file to: {path}")
        with open(path, "w", encoding=encoding) as output_file:
            output_file.write(yaml.dump(self.data, sort_keys=False))



from dataclasses import dataclass
from enum import Enum
from typing import Dict, List, Literal, Optional, Tuple

from cognite.powerops.data_classes.common import AggregationMethod, RetrievalType
from cognite.powerops.data_classes.time_series_mapping import TimeSeriesMapping, TimeSeriesMappingEntry
from cognite.powerops.data_classes.transformation import Transformation, TransformationType


class Auction(str, Enum):
    week = "week"
    weekend = "weekend"


# TODO: Switch to class inheriting from str and Enum, or 3.11 strEnum?
Product = Literal["up", "down"]
Block = Literal["day", "night"]


# ! TODO: what about daylight saving ??
def generate_reserve_schedule(volume: int, n_days: int, night: bool = False) -> Dict[str, int]:
    """Creates the argument for the mapping transformation function for RKOM reserve obligation.
        Some examples:
            {"0": 10, "60": 0}
                --> Obligation of 10MW the first hour, then no obligation
            {"60": 10, "120": 0, "60": 10}
                --> No obligation the first hour (implicit), then 10 MW for one hour,
                    then no obligation for one hour, and finally 10 MW for the rest of the *period*

    Args:
        volume (int): volume in MW
        n_days (int): the number of days to generate the schedule for.
        night (bool, optional): True if the obligation is during the night. Defaults to False.

    Returns:
        Dict[str, int]: The "obligation schedule".
    """
    if volume == 0:
        return {"0": 0}

    res = {}
    minutes_five_hours = 5 * 60
    minutes_day = 24 * 60
    for day in range(n_days):
        midnight = day * minutes_day
        at_5AM = day * minutes_day + minutes_five_hours
        res[f"{midnight}"] = volume if night else 0
        res[f"{at_5AM}"] = 0 if night else volume
    if not night:
        # only needed if last value was != 0 (i.e. during day)
        res[f"{minutes_day * n_days}"] = 0
    return res


@dataclass
class ReserveScenario:
    volume: int
    auction: Auction
    product: Product
    block: Block
    reserve_group: str
    mip_plant_time_series: List[
        Tuple[str, Optional[str]]
    ]  # (plant, mip_flag_time_series) for plants with generators that are in the reserve group
    obligation_external_id: Optional[str] = None

    def __post_init__(self):
        if len(self.mip_plant_time_series) == 0:
            raise ValueError("No `mip_plants` specified!")

    @property
    def obligation_transformations(self) -> List[Transformation]:
        # TODO: move some of this logic
        n_days = 5 if self.auction == "week" else 2
        night = self.block == "night"
        kwargs = generate_reserve_schedule(volume=self.volume, n_days=n_days, night=night)
        if self.obligation_external_id:
            return [Transformation(transformation=TransformationType.DYNAMIC_ADD_FROM_OFFSET, kwargs=kwargs)]
        else:
            return [Transformation(transformation=TransformationType.DYNAMIC_STATIC, kwargs=kwargs)]

    def mip_flag_transformations(self, mip_time_series: Optional[str]) -> List[Transformation]:
        if not mip_time_series:
            # Just run with_mip flag the entire period
            return [Transformation(transformation=TransformationType.STATIC, kwargs={"0": 1})]

        # Run with mip_flag during bid, in addition to what is already set
        n_days = 5 if self.auction == "week" else 2
        n_minutes = n_days * 24 * 60
        kwargs = {0: 1, n_minutes: 0}
        return [
            Transformation(transformation=TransformationType.DYNAMIC_ADD_FROM_OFFSET, kwargs=kwargs),
            Transformation(transformation=TransformationType.TO_BOOL),
        ]

    @property
    def obligation_object_type(self) -> str:
        return "reserve_group"

    @property
    def obligation_attribute_name(self) -> str:
        if self.product == "up":
            return "rr_up_obligation"
        elif self.product == "down":
            return "rr_down_obligation"

    def to_time_series_mapping(self) -> TimeSeriesMapping:
        obligation = TimeSeriesMappingEntry(
            object_type=self.obligation_object_type,
            object_name=self.reserve_group,
            attribute_name=self.obligation_attribute_name,
            time_series_external_id=self.obligation_external_id,
            transformations=self.obligation_transformations,
            retrieve=RetrievalType.RANGE if self.obligation_external_id else None,
        )

        mip_flags = [
            TimeSeriesMappingEntry(
                object_type="plant",
                object_name=plant_name,
                attribute_name="mip_flag",
                time_series_external_id=mip_time_series,
                transformations=self.mip_flag_transformations(mip_time_series),
                retrieve=RetrievalType.RANGE if mip_time_series else None,
                aggregation=AggregationMethod.max,  # TODO: or `AggregationMethod.first`?
            )
            for plant_name, mip_time_series in self.mip_plant_time_series
        ]

        return TimeSeriesMapping(rows=[obligation, *mip_flags])



from __future__ import annotations

from hashlib import md5
from pathlib import Path
from typing import Literal, Optional

from cognite.client.data_classes import FileMetadata
from pydantic import BaseModel

from cognite.powerops.utils.serializer import load_yaml


class ShopFileConfig(BaseModel):
    watercourse_name: str
    path: Path
    cogshop_file_type: Literal[
        "case",
        "model",
        "log",
        "commands",
        "extra_data",
        "water_value_cut_file_reservoir_mapping",
        "water_value_cut_file",
        "module_series",
    ]
    md5_hash: Optional[str] = None

    @property
    def external_id(self) -> str:
        return f"SHOP_{self.watercourse_name}_{self.path.stem}"

    @property
    def metadata(self) -> dict[str, str]:
        shop_type = self.cogshop_file_type
        if shop_type == "model":
            shop_type = "case"
        if self.md5_hash is None:
            raise AttributeError("The md5 hash is not set")
        return {
            "shop:type": shop_type,
            "shop:watercourse": self.watercourse_name,
            "shop:file_name": self.path.stem,
            "md5_hash": self.md5_hash,
        }

    def set_md5_hash(self, file_content: bytes) -> None:
        self.md5_hash = md5(file_content).hexdigest()

    def set_full_path(self, directory: str) -> None:
        self.path = Path(directory) / self.path

    @classmethod
    def from_file_meta(cls, file_meta: FileMetadata) -> "ShopFileConfig":
        cogshop_file_type = file_meta.metadata["shop:type"] if file_meta.metadata["shop:type"] != "case" else "model"
        return cls(
            path=Path(file_meta.metadata.get("shop:file_name") or cogshop_file_type),
            watercourse_name=file_meta.metadata["shop:watercourse"],
            cogshop_file_type=cogshop_file_type,
            md5_hash=file_meta.metadata.get("md5_hash"),
        )


class ShopFileConfigs(BaseModel):
    watercourses_shop: list[ShopFileConfig]

    @classmethod
    def from_yaml(cls, config_dir_path: Path) -> "ShopFileConfigs":
        configs = {}
        for field_name in cls.__fields__:
            if (config_file_path := config_dir_path / f"{field_name}.yaml").exists():
                configs[field_name] = load_yaml(config_file_path, encoding="utf-8")
        return cls(**configs)



from datetime import datetime
from typing import Dict, List, Optional, Union

import pandas as pd

from cognite.powerops.preprocessor import knockoff_logging as logging
from cognite.powerops.preprocessor.utils import datetime_from_str

logger = logging.getLogger(__name__)


def find_resolution_ranges(resolution: Dict[datetime, int]) -> List[tuple]:
    """Finds the start and end of a period with a given resolution. The resolution dictionatry
    is expected to be on the format:
        {<datetime>: <resolution>, <other datetime>: <other resolution>, ...}

    The ranges are stored as tuples in a list where the entries of the tuples are:
        [(<resolution>, <start datetime>, <end datetime>), ...]
    """
    if not resolution:
        return []

    ordered = sorted(list(resolution.items()))  # order by timestamp
    ranges = []
    current_start = ordered[0][0]
    current_resolution = ordered[0][1]
    for i in range(len(ordered)):
        next_resolution = ordered[i + 1][1] if (i < len(ordered) - 1) else None
        next_timestamp = ordered[i + 1][0] if (i < len(ordered) - 1) else None
        # If change in resolution or last element
        if current_resolution != next_resolution or not next_resolution:
            ranges.append((current_resolution, current_start, next_timestamp))
            current_resolution = next_resolution  # type: ignore
            current_start = next_timestamp  # type: ignore
    return sorted(ranges)


def resample(series: pd.Series, resolution: str, method: str) -> pd.Series:
    """Resample series according to the given resolution
    Returns a pandas series representing the resampled datapoints
    """
    resampler = series.resample(resolution, origin=series.index.min())
    if method not in {"sum", "mean", "std", "max", "min", "median", "first", "last"}:
        logger.error(f"{method} is not a supported operation")  # TODO: raise here instead?
        return series
    return resampler.aggregate(func=method)  # type: ignore[return-value]


def _resolution_dict_to_datetime(resolution_dict: Dict[str, int]) -> Dict[datetime, int]:
    """resolution dict might be given on the format {string: int}
    -> but aggregate expects resolution dict to be {datetime: int}
    """
    return {datetime_from_str(k) if isinstance(k, str) else k: v for k, v in resolution_dict.items()}


def resample_and_aggregate(
    datapoints: Union[float, pd.Series],
    method: Optional[str],
    resolution_dict: Dict[str, int],
    is_step: bool = True,
) -> Union[float, pd.Series]:
    """Resample and aggregate datapoints according to the specified resolution in that period"""
    # Only relevant if series
    if not isinstance(datapoints, pd.Series):
        return datapoints

    # Cannot resample and aggregate without a specified method
    if not method:
        return datapoints

    resolutions = _resolution_dict_to_datetime(resolution_dict)
    if min(datapoints.index) < min(resolutions):
        raise ValueError(
            f"Resolution specified from {min(resolutions)} but got datapoints from {min(datapoints.index)}!"
        )

    # Resample and forwardfill or interpolate so that aggregations to higher granularities become correct
    # NOTE: assumes that 1h is the lowest granularity of datapoints
    if is_step:
        series = datapoints.resample("1h").ffill()
    else:
        series = datapoints.resample("1h").interpolate()

    resampled = []
    for resolution, start, end in find_resolution_ranges(resolutions):
        end = end or pd.Timestamp.max  # if no end specified
        mask = (start <= series.index) & (series.index < end)
        sub_series = series.loc[mask]
        sub_resampled = resample(series=sub_series, resolution=f"{int(resolution)}min", method=method)
        resampled.append(sub_resampled)

    return pd.concat(resampled)



import getpass
from typing import Optional, Union

from cognite.client import ClientConfig, CogniteClient
from cognite.client.credentials import OAuthClientCredentials, OAuthDeviceCode
from pydantic import BaseModel, validator

from cognite.powerops.settings import settings


class _CogniteConfig(BaseModel):
    project: str
    cdf_cluster: str
    tenant_id: str
    client_id: str
    client_secret: Optional[str]

    client_name: str = ""
    authority_host_uri: str = "https://login.microsoftonline.com"

    @property
    def base_url(self) -> str:
        return f"https://{self.cdf_cluster}.cognitedata.com/"

    @property
    def scopes(self) -> list[str]:
        return [f"{self.base_url}.default"]

    @property
    def authority_uri(self) -> str:
        return f"{self.authority_host_uri}/{self.tenant_id}"

    @validator("client_name", always=True)
    def replace_none_with_user(cls, value):
        return getpass.getuser() if value is None else value


def get_client_config(
    project: Optional[str] = None,
    cdf_cluster: Optional[str] = None,
    tenant_id: Optional[str] = None,
    client_id: Optional[str] = None,
    client_secret: Optional[str] = None,
    client_name: str = "",
) -> ClientConfig:
    cognite_config = _CogniteConfig(
        project=project or settings.cognite.project,
        cdf_cluster=cdf_cluster or settings.cognite.cdf_cluster,
        tenant_id=tenant_id or settings.cognite.tenant_id,
        client_id=client_id or settings.cognite.client_id,
        client_secret=client_secret or settings.cognite.client_secret,
        client_name=client_name,
    )

    credentials: Union[OAuthClientCredentials, OAuthDeviceCode]
    if cognite_config.client_secret:
        credentials = OAuthClientCredentials(
            token_url=f"https://login.microsoftonline.com/{cognite_config.tenant_id}/oauth2/v2.0/token",
            client_id=cognite_config.client_id,
            client_secret=cognite_config.client_secret,
            scopes=cognite_config.scopes,
        )
    elif cognite_config.tenant_id:
        credentials = OAuthDeviceCode(
            authority_url=cognite_config.authority_uri,
            client_id=cognite_config.client_id,
            scopes=cognite_config.scopes,
        )
    else:
        raise ValueError("Missing authentication details for CDF")
    return ClientConfig(
        client_name=cognite_config.client_name,
        project=cognite_config.project,
        base_url=cognite_config.base_url,
        credentials=credentials,
    )


def get_cognite_client(
    project: Optional[str] = None,
    cdf_cluster: Optional[str] = None,
    tenant_id: Optional[str] = None,
    client_id: Optional[str] = None,
    client_secret: Optional[str] = None,
    client_name: str = "",
) -> CogniteClient:
    return CogniteClient(get_client_config(project, cdf_cluster, tenant_id, client_id, client_secret, client_name))
